
package org.xdat.chart;

import java.awt.Color;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import org.xdat.data.DataSheet;
import org.xdat.data.Design;

/**
 * Reprezentation of category of axis.
 * Each axis of parallel sets chart is divided into a specified number of
 * categories. The category covers the part of the axis that is defined by
 * condition.
 *
 * @author Martin Janda
 */
public class AxisCategory implements Serializable {
    
    /** Vertical space between categories on the axis */
    private static int Y_OFFSET = 10;
    
    /** Condition that determines which designs belong to the category */
    private AxisCategoryCondition condition;
    
    /** Axis of this category */
    private ParallelSetsAxis axis;
    
    /** Position of upper edge relative to axis in pixels */
    private int yUpPosition;
    
    /** Temporary height of output sets for rendering */
    private float outSetsHeight;
    
    /** Temporary height of input sets for rendering */
    private float inSetsHeight;
    
    /** Height of category to height of axis ratio */
    private double heightRatio;
    
    /** Height of category in pixels */
    private int height;
    
    /** Label of this category */
    private String label;

    /** Designs that belong to this category */
    private List<Design> designs;
    
    /** Color of this category */
    private Color color;
    
    private List<InsideDesignSet> insideSets;
    
    private boolean selected;
    public boolean active = true;
    
    private List<DesignSet> inputSets;
    private List<DesignSet> outputSets;
    
    private int order;
    
    public AxisCategory(AxisCategoryCondition condition, double heightRatio, List<Design> designs, ParallelSetsAxis axis) {
        this.condition = condition;
        this.heightRatio = heightRatio;
        this.designs = designs;
        this.axis = axis;
        yUpPosition = 0;
        height = 0;
        inputSets = new ArrayList<>();
        outputSets = new ArrayList<>();
        insideSets = new ArrayList<>();
        outSetsHeight = 0;
        inSetsHeight = 0;
        order = 0;
        
        generateLabel();
    }

    public void setCondition(AxisCategoryCondition condition) {
        this.condition = condition;
    }

    public int getOrder() {
        return order;
    }

    public void setOrder(int order) {
        this.order = order;
    }

    public boolean isSelected() {
        return selected;
    }

    public void setSelected(boolean selected) {
        this.selected = selected;
    }

    public List<InsideDesignSet> getInsideSets() {
        return insideSets;
    }

    public void setInsideSets(List<InsideDesignSet> insideSets) {
        this.insideSets = insideSets;
    }
    
    public Color getColor() {
        return color;
    }

    public void setColor(Color color) {
        this.color = color;
    }
    
    public void setAlpha(int alpha) {
        color = new Color(color.getRed(), color.getGreen(), color.getBlue(), alpha);
    }

    public float getOutSetsHeight() {
        return outSetsHeight;
    }

    public void setOutSetsHeight(float outSetsHeight) {
        this.outSetsHeight = outSetsHeight;
    }

    public float getInSetsHeight() {
        return inSetsHeight;
    }

    public void setInSetsHeight(float inSetsHeight) {
        this.inSetsHeight = inSetsHeight;
    }

    public List<DesignSet> getInputSets() {
        return inputSets;
    }

    public void setInputSets(List<DesignSet> inputSets) {
        this.inputSets = inputSets;
    }

    public List<DesignSet> getOutputSets() {
        return outputSets;
    }

    public void setOutputSets(List<DesignSet> outputSets) {
        this.outputSets = outputSets;
    }

    public int getHeight() {
        return height;
    }

    public void setHeight(int height) {
        this.height = height;
    }

    public int getyUpPosition() {
        return yUpPosition;
    }

    public void setyUpPosition(int yUpPosition) {
        this.yUpPosition = yUpPosition;
    }

    public static int getYOffset() {
        return Y_OFFSET;
    }

    public List<Design> getDesigns() {
        return designs;
    }

    public void setDesigns(List<Design> designs) {
        this.designs = designs;
    }

    public AxisCategoryCondition getCondition() {
        return condition;
    }

    public double getHeightRatio() {
        return heightRatio;
    }

    public void setHeightRatio(double heightRatio) {
        this.heightRatio = heightRatio;
    }

    public String getLabel() {
        return label;
    }

    public void setLabel(String label) {
        this.label = label;
    }
    
    public void generateLabel(){
        label = condition.getLabel();
    }

    public ParallelSetsAxis getAxis() {
        return axis;
    }

    public void setAxis(ParallelSetsAxis axis) {
        this.axis = axis;
    }
    
    public String toString(){
        return label;
    }
    
    public boolean containsPoint(int y) {
        if(y > getyUpPosition() && y < getyUpPosition() + getHeight()) {
           return true;
        }
        return false;
    }

    /**
     * @return the active
     */
    public boolean isActive() {
        return active;
    }

    /**
     * @param active the active to set
     */
    public void setActive(boolean active) {
        DataSheet dataSheet = this.getAxis().getDataSheet();
        long designsChanged = 0;
        for(Design d : getDesigns()) {
            if(d.isActive() != active) {
                designsChanged++;
                d.setActive(active);
            }
        }
        long inactiveDesigns = dataSheet.getInactiveDesigns();
        if(active) {
            inactiveDesigns -= designsChanged;
            //axis.inactiveCategories.remove(getLabel());
        }
        else {
            inactiveDesigns += designsChanged;
            //axis.inactiveCategories.add(getLabel());
        }
        dataSheet.setInactiveDesigns(inactiveDesigns);
        this.active = active;
    }
}
