/*
 *  Copyright 2014, Enguerrand de Rochefort
 * 
 * This file is part of xdat.
 *
 * xdat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * xdat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with xdat.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.xdat.chart;

import java.io.Serializable;
import java.util.Vector;

import javax.swing.ProgressMonitor;

import org.xdat.Main;
import org.xdat.data.DataSheet;

/**
 * A serializable representation of all relevant settings for a Parallel
 * coordinates chart which is displayed on a ChartFrame.
 * <p>
 * The data that was imported by the user can be displayed on a Parallel
 * coordinates Chart. The Parallel coordinates Chart is built by as many
 * vertical Axes as Parameters are present in the underlying DataSheet and each
 * Axis represents one of these Parameters.
 * <p>
 * The Designs (or, in other words, rows of the data table) are represented by
 * lines that connect points on the Axes. Each Design's line crosses each Axis
 * exactly at the ordinate that corresponds to the value for the respective
 * Parameter in the Design. This allows for displaying the whole DataSheet in
 * just one Chart, irrespective of how many dimensions it has.
 * <p>
 * The Chart also provides interactivity through a pair of draggable Filters
 * that are present on each Axis. Depending on the positions of these filters,
 * certain Designs are filtered from the display. This means that they are
 * either displayed in a different color or hidden completely.
 * 
 * 
 * @see org.xdat.gui.frames.ChartFrame
 * @see Axis
 * @see Filter
 * @see org.xdat.data.Parameter
 * @see org.xdat.data.DataSheet
 * @see org.xdat.data.Design
 */
public class ParallelCoordinatesChart extends ParallelChart implements Serializable {

	/** Flag to enable debug message printing for this class. */
	static final boolean printLog = false;

	/**
	 * The axes on this Chart.
	 * <p>
	 * For each Parameter in the DataSheet there is exactly one Axis.
	 * */
	private Vector<Axis> axes = new Vector<Axis>(0, 1);

	/**
	 * Instantiates a new parallel coordinates chart.
	 * 
	 * @param dataSheet
	 *            the data sheet
	 * @param progressMonitor 
	 * 			the progress monitor
	 * @param id 
	 * 				the id
	 */
	public ParallelCoordinatesChart(DataSheet dataSheet, ProgressMonitor progressMonitor, int id) {
            super(dataSheet, progressMonitor, id);
           
		for (int i = 0; i < dataSheet.getParameterCount() && !progressMonitor.isCanceled(); i++) {
			// log("constructor: Creating axis "+dataSheet.getParameter(i).getName());
			Axis newAxis = new Axis(dataSheet, this, dataSheet.getParameter(i));
			this.addAxis(newAxis);
			progressMonitor.setProgress(i);
		}

		if (!progressMonitor.isCanceled()) {
			progressMonitor.setNote("Building Filters...");
			progressMonitor.setProgress(0);
			// log("constructor: axes created. Creating filters...");
			for (int i = 0; i < dataSheet.getParameterCount() && !progressMonitor.isCanceled(); i++) {
				this.axes.get(i).addFilters();
				progressMonitor.setProgress(i);
			}
			// log("constructor: filters created. ");
		}

	}

	/**
	 * Determines the width of this Chart.
	 * 
	 * @return the width of this Chart
	 */
	public int getWidth() {
		int width = 0;
		if (this.getAxis(0).isActive()) {
			width = width + (int) (0.5 * this.getAxis(0).getWidth());
		}
		for (int i = 1; i < this.getAxisCount(); i++) {
			if (this.getAxis(i).isActive()) {
				width = width + (int) (this.getAxis(i).getWidth());
			}
		}
		return width;
	}

	/**
	 * Gets an Axis by its index.
	 * 
	 * @param index
	 *            the index
	 * @return the Axis with index index
	 */
	public Axis getAxis(int index) {

		return axes.get(index);
	}

	/**
	 * Gets the axis count.
	 * 
	 * @return the axis count
	 */
	public int getAxisCount() {
		return axes.size();
	}

	/**
	 * Adds the axis.
	 * 
	 * @param axis
	 *            the axis
	 */
	public void addAxis(Axis axis) {
		this.axes.add(axis);
	}

	/**
	 * Adds an Axis at the position index.
	 * 
	 * @param index
	 *            the index where the Axis should be added
	 * @param axis
	 *            the Axis to be added
	 */
	public void addAxis(int index, Axis axis) {
		this.axes.add(index, axis);
	}

	/**
	 * Removes the axis with index index.
	 * 
	 * @param index
	 *            the index of the Axis to be removed
	 */
	public void removeAxis(int index) {
		this.axes.remove(index);
	}

	/**
	 * Function to reorder the axes in the chart
	 * 
	 * @param oldIndex
	 *            the index of the axis to be moved
	 * @param newIndex
	 *            the target index for the axis to be moved
	 */
	public void moveAxis(int oldIndex, int newIndex) {
		log("moveAxis called with arguments " + oldIndex + " and " + newIndex);
		Axis axis = this.axes.remove(oldIndex);
		this.axes.insertElementAt(axis, newIndex);
	}

        @Override
        public String getTitle() {
                return "Parallel Coordinates Chart " + this.getID();
        }
        
	/**
	 * Prints debug information to stdout when printLog is set to true.
	 * 
	 * @param message
	 *            the message
	 */
	private void log(String message) {
		if (ParallelCoordinatesChart.printLog && Main.isLoggingEnabled()) {
			System.out.println(this.getClass().getName() + "." + message);
		}
	}

}
