
package org.xdat.chart;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.font.FontRenderContext;
import java.awt.font.TextAttribute;
import java.util.*;

import org.xdat.UserPreferences;
import org.xdat.data.DataSheet;
import org.xdat.data.Design;
import org.xdat.data.Parameter;

/**
 * Representation of axis of parallel sets chart.
 * Axis is used to represent a Parameter. Each axis is divided into several
 * categories according to parameter values.
 *
 * @author Martin Janda
 */
public class ParallelSetsAxis extends Axis {

    /** Default value of axis width */
    public static final int DEFAULT_BOX_WIDTH = 50;

    /** Default value of coordinates axis width */
    public static final int BOX_WIDTH_WITH_COORDINATES = 150;

    /** Scale of agregated axis */
    public static final float AGGREGATED_AXIS_SCALE = 0.9f;

    /** Height of axis label */
    public static final int LABEL_HEIGHT = 10;

    /** Vertical distance of axis label */
    public static final int LABEL_VERTICAL_DISTANCE = 10;
    
    public transient FontRenderContext fontContext;
    
    protected boolean selected;
    protected boolean expanded; 
    protected boolean coordinatesAxis;
    protected boolean inverted;
    private boolean userSorted = false;
    protected HashMap<String, Integer> categoriesOrder = new HashMap<>();
    HashSet<String> selectedCategories = new HashSet<>();
    HashSet<String> inactiveCategories = new HashSet<>();
    
    protected ParallelSetsChart parallelSetsChart;
    
    protected List<AxisCategory> categories = new ArrayList<>();

    protected Font labelFont;
    protected Font labelFontSelected;
    protected Font ticFont;

    private int yLabelOffset;
    
    protected ParallelSetsAxis() {}
    
    public ParallelSetsAxis(DataSheet dataSheet, ParallelSetsChart parallelSetsChart, Parameter parameter) {
        super(dataSheet, parallelSetsChart, parameter);

        yLabelOffset = 0;
        selected = false;
        expanded = false;
        this.parallelSetsChart = parallelSetsChart;
        coordinatesAxis = false;
        inverted = false;
        
        int labelFontSize = UserPreferences.getInstance().getParallelCoordinatesAxisLabelFontSize();
        labelFont = new Font("SansSerif", Font.PLAIN, labelFontSize);
        Map<TextAttribute, Integer> fontAttributes = new HashMap<>();
        fontAttributes.put(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
        labelFontSelected = labelFont.deriveFont(fontAttributes);
        int ticFontSize = UserPreferences.getInstance().getParallelCoordinatesAxisTicLabelFontSize();
        ticFont = new Font("SansSerif", Font.PLAIN, ticFontSize);
    }

    public Font getLabelFont() {
        return labelFont;
    }

    public void setLabelFont(Font labelFont) {
        this.labelFont = labelFont;
    }

    public boolean isInverted() {
        return inverted;
    }

    public void setInverted(boolean inverted) {
        this.inverted = inverted;
    }
    
    public boolean isCoordinatesAxis() {
        return coordinatesAxis;
    }

    public void setCoordinatesAxis(boolean coordinatesAxis) {
        if(coordinatesAxis){
            showCoordinatesAxis();
        }else{
            hideCoordinatesAxis();
        }
    }
    
    private void showCoordinatesAxis(){
        if(!getParameter().isNumeric()) return; // only numeric axes can show coordinates axis
        coordinatesAxis = true;

        updateCategoriesHeightRatio();
    }
    
    private void hideCoordinatesAxis(){
        coordinatesAxis = false;

        updateCategoriesHeightRatio();
    }
    
    public void updateCategoriesHeightRatio(){
        if(coordinatesAxis && parallelSetsChart.getCoordAxisScaleMode() == ParallelSetsChart.COORD_AXIS_SCALING_CATEGORIES){
            for (AxisCategory category : categories) {
                if (category.getCondition().getClass().equals(AxisCategoryStringCondition.class)) {
                    category.setHeightRatio(0.005);
                } else {
                    AxisCategoryRangeCondition condition = (AxisCategoryRangeCondition) category.getCondition();
                    double ratio = (condition.getMax() - condition.getMin()) / getRange();
                    category.setHeightRatio(ratio);
                }
            }
        }else{
            double designsCount = (double) (parallelSetsChart.getDataSheet().getDesignCount() - parallelSetsChart.getDataSheet().getInactiveDesigns());
            for (AxisCategory category : categories) {
                double ratio = (double) category.getDesigns().size() / designsCount;
                category.setHeightRatio(ratio);
            }
        }
    }

    public boolean isSelected() {
        return selected;
    }

    public void setSelected(boolean selected) {
        this.selected = selected;
    }
    
    public List<AxisCategory> getCategories(){
        return categories;
    }

    public void setCategories(List<AxisCategory> categories) {
        this.categories = categories;
    }
    
    public int getNumberOfAllCategories() {
        return categories.size();
    }
    
    public int getNumberOfActiveCategories(){
        int number = 0;
        for(AxisCategory c : categories) {
            if(c.isActive()) number++;
        }
        return number;
    }
    
    public AxisCategory getCategoryFromAll(int index){
        return categories.get(index);
    }
    
    public AxisCategory getCategoryFromActive(int index){
        int i = 0;
        for(AxisCategory c : getCategories()) {
            if(c.isActive()) {
                if(i == index) return c;
                i++;
            }
            
        }
        return null;
    }

    public int getBoxWidth() {
        return isCoordinatesAxis() ? parallelSetsChart.getCoordinateAxisBoxWidth() : parallelSetsChart.getStandardAxisBoxWidth();
    }

    @Override
    public void setActive(boolean active) {
        super.setActive(active);
        if(!active && parallelSetsChart.getSeparatingAxis() == this){
            ParallelSetsAxis firstActiveAxis = this;
            for(int i = 0; i < parallelSetsChart.getAxisCount(); i++){
                ParallelSetsAxis axis = parallelSetsChart.getAxis(i);
                if(axis.isActive()){
                    firstActiveAxis = axis;
                    break;
                }
            }
            parallelSetsChart.setSeparatingAxis(firstActiveAxis);
        }
        parallelSetsChart.buildDesignSets();
    }
    
    public boolean isExpanded() {
        return expanded;
    }
    
    public void setExpanded(boolean expanded) {
        this.expanded = expanded;
    }
    
    public void update(int xPosition) {
        double yPosition = getChart().getAxisTopPos();
        int axisHeight = getChart().getAxisHeight();
        
        Parameter p = getParameter();
        if(p != null && p.isAggregated()) {
            int aggregatedAxisHeihgt = (int)(axisHeight * AGGREGATED_AXIS_SCALE);
            yPosition += (axisHeight - aggregatedAxisHeihgt)/2 + 15;
            axisHeight = aggregatedAxisHeihgt;
        }

        setPosition(new Point(xPosition, (int) yPosition));
        
        int numberOfCategories = getNumberOfActiveCategories() - 1;
        if(isCoordinatesAxis()) {
            numberOfCategories = getNumberOfAllCategories() - 1;
        }

        double usableHeight = axisHeight -  numberOfCategories* AxisCategory.getYOffset();
        
        // update categories
        for (AxisCategory category : getCategories()) {
            if(!category.isActive() && !isCoordinatesAxis()) continue;
            category.setyUpPosition((int) yPosition);
            
            double height = usableHeight * category.getHeightRatio();
            category.setHeight((int) height);

            yPosition += height;
            yPosition += AxisCategory.getYOffset();
        }
    } 
    
    @Override
    public ParallelChart getChart() {
        return parallelSetsChart;
    }

    public int getYLabelOffset() {
        return yLabelOffset;
    }

    public void setYLabelOffset(int yLabelOffset) {
        this.yLabelOffset = yLabelOffset;
    }

    public void drawAxis(Graphics2D g) {
        // Axis label
        if (chart != null && chart.isVerticallyOffsetAxisLabels()) {
            yLabelOffset = (((ParallelSetsChart)chart).getDisplayAxes().indexOf(this) % 2) * (LABEL_HEIGHT + LABEL_VERTICAL_DISTANCE);
        }

        String axisLabel = getName();
        g.setFont(labelFont);
        fontContext = g.getFontRenderContext();
        if (parallelSetsChart.getSeparatingAxis() == this) { // Selected axis label
            g.setFont(labelFontSelected);
        }
        g.setColor(Color.black);
        g.drawString(axisLabel,
                getPosition().x,
                getPosition().y - ParallelSetsAxis.LABEL_HEIGHT - yLabelOffset);

        // Sort button
        if (parallelSetsChart.isSortButtonOn() && getParameter() != null && getParameter().isNumeric()) {
            String arrow = "↑";
            if (inverted) arrow = "↓";
            g.drawString(arrow, getPosition().x - 15, getPosition().y - ParallelSetsAxis.LABEL_HEIGHT - yLabelOffset);
        }

        // Axis
        if (isCoordinatesAxis()) {
            drawCoordinateAxis(g);
        } else {
            drawCategoryAxis(g);
        }
    }

    protected void drawCategoryAxis(Graphics2D g) {
        // Categories
        List<AxisCategory> categories = getCategories();
        for (AxisCategory category : categories) {
            if (!category.isActive()) continue;
            // Selected category
            if (category.isSelected()) {
                g.setStroke(new BasicStroke(2));
            }
            g.setColor(Color.black);
            g.drawRect(getPosition().x, category.getyUpPosition(), getBoxWidth(), category.getHeight());
            g.setStroke(new BasicStroke());

            UserPreferences pref = UserPreferences.getInstance();
            int fontSize = pref.getParallelCoordinatesAxisTicLabelFontSize();
            g.setColor(Color.black);

            //Font font = new Font("SansSerif", Font.PLAIN, fontSize);
            g.setFont(ticFont);
            //Rectangle2D bounds = font.getStringBounds("x", fontContext);
            float textHeight = 7.5f;
            String cLabel = category.getLabel();
            String[] rows = cLabel.split("\\n");
            float labelX = getPosition().x + getBoxWidth() + 5;
            float labelY = (float) (category.getyUpPosition() + category.getHeight() / 2f + textHeight / 2f - textHeight * (rows.length - 1) / 2f);
            for (String row : rows) {
                g.drawString(row, labelX, labelY);
                labelY += textHeight;
            }
        }
    }


    protected void drawCoordinateAxis(Graphics2D g) {
        if (parallelSetsChart.getCoordAxisScaleMode() == ParallelSetsChart.COORD_AXIS_SCALING_AXIS) {

            int xPosition = getPosition().x + getBoxWidth() / 2;
            int yPosition = getPosition().y;


            // draw categories
            List<AxisCategory> categories = getCategories();
            for (AxisCategory category : categories) {
                if (!category.isActive()) continue;
                if (!category.getCondition().getClass().equals(AxisCategoryRangeCondition.class)) {
                    continue;
                }
                AxisCategoryRangeCondition condition = (AxisCategoryRangeCondition) category.getCondition();

                // draw coordinates axis
                g.drawLine(xPosition, category.getyUpPosition(), xPosition, category.getyUpPosition() + category.getHeight());

                // draw categories
                double topValue = condition.getMax();
                double bottomValue = condition.getMin();

                double topRatio = (topValue - getMin()) / getRange();
                int topPosRelToBottom = (int) (parallelSetsChart.getAxisHeight() * topRatio);
                double bottomRatio = (bottomValue - getMin()) / getRange();
                int bottomPosRelToBottom = (int) (parallelSetsChart.getAxisHeight() * bottomRatio);

                int yTopOnAxis = yPosition + parallelSetsChart.getAxisHeight() - topPosRelToBottom;
                int yBottomOnAxis = yPosition + parallelSetsChart.getAxisHeight() - bottomPosRelToBottom;


                // draw tics
                int ticSize = getTicLength();
                int ticCount = (int) ((double) getTicCount() * category.getHeightRatio());

                double ticSpacing;
                if (ticCount > 1)
                    ticSpacing = category.getHeight() / ((double) (ticCount - 1));
                else
                    ticSpacing = 0;
                double categoryRange = topValue - bottomValue;
                double ticValueDifference = categoryRange / ((double) (ticCount - 1));
                for (int ticID = 0; ticID < ticCount; ticID++) {
                    int currentTicYPos;
                    if (isInverted())
                        currentTicYPos = category.getyUpPosition() + category.getHeight() - (int) (ticID * ticSpacing);
                    else
                        currentTicYPos = category.getyUpPosition() + (int) (ticID * ticSpacing);
                    g.setColor(Color.black);
                    if (ticCount > 1)
                        g.drawLine(xPosition, currentTicYPos, xPosition + ticSize, currentTicYPos);
                    else
                        g.drawLine(xPosition, category.getyUpPosition() + (int) (category.getHeight() / 2), xPosition + ticSize, category.getyUpPosition() + (int) (category.getHeight() / 2));

                    g.setColor(Color.black);

                    String ticLabel;
                    g.setFont(new Font("SansSerif", Font.PLAIN, getTicLabelFontSize()));
                    //if (currentAxis.getParameter().isNumeric()) {
                    Double ticValue;
                    if (ticCount > 1) {
                        ticValue = topValue - ticValueDifference * ticID;
                        ticLabel = String.format(getTicLabelFormat(), ticValue);
                        g.drawString(ticLabel, xPosition + ticSize + 7, currentTicYPos + (int) (0.5 * getTicLabelFontSize()));
                    } else {
                        ticValue = topValue;
                        ticLabel = String.format(getTicLabelFormat(), ticValue);
                        g.drawString(ticLabel, xPosition + 2 * ticSize, category.getyUpPosition() + ((int) (category.getHeight() / 2)) + (int) (0.5 * getTicLabelFontSize()));
                    }
                }


                int[] pointsX = {
                        getPosition().x,
                        //xPosition,
                        getPosition().x + getBoxWidth(),
                        getPosition().x + getBoxWidth(),
                        //xPosition,
                        getPosition().x
                };
                int[] pointsY = {
                        category.getyUpPosition(),
                        //yTopOnAxis,
                        category.getyUpPosition(),
                        category.getyUpPosition() + category.getHeight(),
                        //yBottomOnAxis,
                        category.getyUpPosition() + category.getHeight()
                };

                // Selected category
                if (category.isSelected()) {
                    g.setStroke(new BasicStroke(2));
                }
                g.setColor(Color.black);
                g.drawPolygon(pointsX, pointsY, 4);
                g.setStroke(new BasicStroke());

                int labelX = getPosition().x + getBoxWidth() + 5;
                int labelY = category.getyUpPosition() + (category.getHeight() / 2);
                g.setColor(Color.black);
                g.setFont(new Font("SansSerif", Font.PLAIN, getTicLabelFontSize()));
                g.drawString(category.getLabel(), labelX, labelY);
            }
        } else {
            // draw coordinates axis
            int xPosition = getPosition().x + getBoxWidth() / 2;
            int yPosition = getPosition().y;

            g.drawLine(xPosition, yPosition, xPosition, yPosition + parallelSetsChart.getAxisHeight());

            // draw tics
            int ticSize = getTicLength();
            int ticCount = getTicCount();
            double ticSpacing;
            if (ticCount > 1)
                ticSpacing = parallelSetsChart.getAxisHeight() / ((double) (ticCount - 1));
            else
                ticSpacing = 0;
            double axisRange = getRange();
            double ticValueDifference = axisRange / ((double) (ticCount - 1));
            for (int ticID = 0; ticID < ticCount; ticID++) {
                int currentTicYPos;
                if (isInverted())
                    currentTicYPos = yPosition + parallelSetsChart.getAxisHeight() - (int) (ticID * ticSpacing);
                else
                    currentTicYPos = yPosition + (int) (ticID * ticSpacing);
                g.setColor(Color.black);
                if (ticCount > 1)
                    g.drawLine(xPosition, currentTicYPos, xPosition + ticSize, currentTicYPos);
                else
                    g.drawLine(xPosition, yPosition + (int) (parallelSetsChart.getAxisHeight() / 2), xPosition + ticSize, yPosition + (int) (parallelSetsChart.getAxisHeight() / 2));

                g.setColor(Color.black);

                String ticLabel;
                g.setFont(new Font("SansSerif", Font.PLAIN, getTicLabelFontSize()));
                if (getParameter().isNumeric()) {
                    Double ticValue;
                    if (ticCount > 1) {
                        ticValue = getMax() - ticValueDifference * ticID;
                        ticLabel = String.format(getTicLabelFormat(), ticValue);
                        g.drawString(ticLabel, xPosition + ticSize + 7, currentTicYPos + (int) (0.5 * getTicLabelFontSize()));
                    } else {
                        ticValue = getMax();
                        ticLabel = String.format(getTicLabelFormat(), ticValue);
                        g.drawString(ticLabel, xPosition + 2 * ticSize, yPosition + ((int) (parallelSetsChart.getAxisHeight() / 2)) + (int) (0.5 * getTicLabelFontSize()));
                    }

                } else {
                    if (ticCount > 1) {
                        ticLabel = getParameter().getStringValueOf(getMax() - ticValueDifference * ticID);
                        g.drawString(ticLabel, xPosition + 2 * ticSize, currentTicYPos + (int) (0.5 * getTicLabelFontSize()));
                    } else {
                        ticLabel = getParameter().getStringValueOf(getMax());
                        g.drawString(ticLabel, xPosition + 2 * ticSize, yPosition + ((int) (parallelSetsChart.getAxisHeight() / 2)) + (int) (0.5 * getTicLabelFontSize()));
                    }
                }
            }

            // draw categories
            List<AxisCategory> categories = getCategories();
            for (AxisCategory category : categories) {
                if (!category.isActive()) continue;
                if (!category.getCondition().getClass().equals(AxisCategoryRangeCondition.class)) {
                    continue;
                }
                AxisCategoryRangeCondition condition = (AxisCategoryRangeCondition) category.getCondition();

                double topValue = condition.getMax();
                double bottomValue = condition.getMin();

                double topRatio = (topValue - getMin()) / getRange();
                int topPosRelToBottom = (int) (parallelSetsChart.getAxisHeight() * topRatio);
                double bottomRatio = (bottomValue - getMin()) / getRange();
                int bottomPosRelToBottom = (int) (parallelSetsChart.getAxisHeight() * bottomRatio);

                int yTopOnAxis;
                int yBottomOnAxis;
                if (isInverted()) {
                    yTopOnAxis = yPosition + bottomPosRelToBottom;
                    yBottomOnAxis = yPosition + topPosRelToBottom;
                } else {
                    yTopOnAxis = yPosition + parallelSetsChart.getAxisHeight() - topPosRelToBottom;
                    yBottomOnAxis = yPosition + parallelSetsChart.getAxisHeight() - bottomPosRelToBottom;
                }

                int[] pointsX = {
                        getPosition().x,
                        xPosition,
                        getPosition().x + getBoxWidth(),
                        getPosition().x + getBoxWidth(),
                        xPosition,
                        getPosition().x
                };
                int[] pointsY = {
                        category.getyUpPosition(),
                        yTopOnAxis,
                        category.getyUpPosition(),
                        category.getyUpPosition() + category.getHeight(),
                        yBottomOnAxis,
                        category.getyUpPosition() + category.getHeight()
                };

                // Selected category
                if (category.isSelected()) {
                    g.setStroke(new BasicStroke(2));
                }
                g.setColor(Color.black);
                g.drawPolygon(pointsX, pointsY, 6);
                g.setStroke(new BasicStroke());

                int labelX = getPosition().x + getBoxWidth() + 5;
                int labelY = category.getyUpPosition() + (category.getHeight() / 2);
                g.setColor(Color.black);
                g.setFont(new Font("SansSerif", Font.PLAIN, getTicLabelFontSize()));
                g.drawString(category.getLabel(), labelX, labelY);
            }
        }
    }

    public boolean containsPoint(int x, int y) {
        if (x > getPosition().x && x < getPosition().x + getBoxWidth()) {
            if (y > parallelSetsChart.getAxisTopPos() && y < parallelSetsChart.getAxisTopPos() + parallelSetsChart.getAxisHeight()) {
                return true;
            }
        }
        return false;
    }

    /**
     * @return the userSorted
     */
    public boolean isUserSorted() {
        return userSorted;
    }

    /**
     * @param userSorted the userSorted to set
     */
    public void setUserSorted(boolean userSorted) {
        this.userSorted = userSorted;
        if(userSorted) {
            int index = 0;
            for(AxisCategory cat : categories) {
                categoriesOrder.put(cat.getLabel(), index);
                index++;
            }
        }
    }
    
    public int getCategoryOrder(AxisCategory cat) {
        Integer order = categoriesOrder.get(cat.getLabel());
        if(order != null) return order;
        else return Integer.MAX_VALUE;
    }

    public AxisCategory getCategoryInDragZone(int y) {
        if(!isActive()) return null;
        AxisCategory previousCategory = null;
        AxisCategory nextCategory = null;
        for(int i = 0; i < getNumberOfActiveCategories(); i++){
            if (i < getNumberOfActiveCategories() - 1) {
                nextCategory = getCategoryFromActive(i+1);
                if (nextCategory != null && !nextCategory.isActive()) nextCategory = null;
            }
            AxisCategory category = getCategoryFromActive(i);
            if(category == null || !category.isActive()) continue;
            int upSpace = getSpaceBelongToCategory(category, previousCategory);
            int downSpace = getSpaceBelongToCategory(category, nextCategory);
            if(y >= category.getyUpPosition() - upSpace && y < category.getyUpPosition() + category.getHeight() + downSpace){
                return category;
            }
            previousCategory = category;
        }
        return null;
    }

    private int getSpaceBelongToCategory(AxisCategory thisCategory, AxisCategory otherCategory) {
        if (thisCategory == null || otherCategory == null) return 0;
        double ratio = (double) otherCategory.getHeight() / (double) (thisCategory.getHeight() + otherCategory.getHeight());
        return (int)(AxisCategory.getYOffset() * ratio);
    }

    public void moveCategory(AxisCategory category, int newIndex) {
        List<AxisCategory> categories = getCategories();
        categories.remove(category);
        categories.add(newIndex, category);
        setUserSorted(true);
    }

    public void updateCategories(int coordAxisScaleMode) {
        DataSheet dataSheet = getDataSheet();
        Parameter parameter = getParameter();

        // Update the categories
        double heightSum = 0;
        for(AxisCategory cat : getCategories()) {
            AxisCategoryCondition c = cat.getCondition();
            List<Design> designsInCat = cat.getDesigns();
//            if(!cat.isActive()) {
//                //cat.setHeightRatio(0);
//                continue;
//            }
            if(designsInCat.isEmpty()) {
                cat.setActive(false);
            }
            double designsCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
            double height = 0;
            if(parameter != null && parameter.isNumeric() && isCoordinatesAxis() && coordAxisScaleMode == ParallelSetsChart.COORD_AXIS_SCALING_CATEGORIES){
                if (c.getClass() == AxisCategoryStringCondition.class) {
                    height = 0.005;
                } else {
                    AxisCategoryRangeCondition rangeCondition = (AxisCategoryRangeCondition) c;
                    height = (rangeCondition.getMax() - rangeCondition.getMin()) / getRange();
                }
            }
            else {
                height = (double) designsInCat.size() / designsCount;
            }
            heightSum += height;
            cat.setHeightRatio(height);
        }
        if(heightSum != 1.0) {
            System.out.println("Height sumfor axis " + getName() + "is " + heightSum);
        }
    }

    public void sortCategories() {
        Parameter parameter = getParameter();
        if(parameter != null && parameter.isNumeric()) {
            Collections.sort(getCategories(), new Comparator<AxisCategory>() {
                @Override
                public int compare(AxisCategory ac1, AxisCategory ac2) {
                    AxisCategoryCondition c1 = ac1.getCondition();
                    AxisCategoryCondition c2 = ac2.getCondition();
                    Double value1;
                    Double value2;
                    if(c1.getClass().equals(AxisCategoryRangeCondition.class)){
                        value1 = ((AxisCategoryRangeCondition) c1).getMin();
                    }else{
                        value1 = Double.NEGATIVE_INFINITY;
                    }
                    if(c2.getClass().equals(AxisCategoryRangeCondition.class)){
                        value2 = ((AxisCategoryRangeCondition) c2).getMin();
                    }else{
                        value2 = Double.NEGATIVE_INFINITY;
                    }
                    if(isInverted()) return value1.compareTo(value2);
                    return value2.compareTo(value1);

                }
            });
        } else {
            //sort by user defined order if it exist
            if(isUserSorted()) {
                Collections.sort(getCategories(), new Comparator<AxisCategory>() {
                    @Override
                    public int compare(AxisCategory ac1, AxisCategory ac2) {
                        return getCategoryOrder(ac1) - getCategoryOrder(ac2);
                    }
                });
            } else {
                Collections.sort(getCategories(), new Comparator<AxisCategory>() {
                    @Override
                    public int compare(AxisCategory ac1, AxisCategory ac2) {
                        // sort by order code if is defined
                        if (parameter != null && !ac1.getDesigns().isEmpty() && !ac2.getDesigns().isEmpty()) {
                            Integer order1 = ac1.getDesigns().get(0).getValueOrder().get(parameter);
                            Integer order2 = ac2.getDesigns().get(0).getValueOrder().get(parameter);
                            if (order1 != null && order2 != null) {
                                return order1.compareTo(order2);
                            }
                        }
                        // if order code is not defined, sort alphabetically
                        AxisCategoryStringCondition c1 = (AxisCategoryStringCondition) ac1.getCondition();
                        AxisCategoryStringCondition c2 = (AxisCategoryStringCondition) ac2.getCondition();
                        String s1 = c1.getString();
                        String s2 = c2.getString();
                        return s2.compareTo(s1);
                    }
                });
            }
        }

        for(int i = 0; i < getCategories().size(); i++){
            getCategories().get(i).setOrder(i);
        }
    }

    @Override
    public int getWidth() {
        return parallelSetsChart.getAllAxesWidth();
    }
}
