
package org.xdat.chart;

import java.awt.Color;
import java.awt.Dimension;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import javax.swing.ProgressMonitor;
import org.xdat.Main;
import org.xdat.UserPreferences;
import org.xdat.data.DataSheet;
import org.xdat.data.Design;
import org.xdat.data.Parameter;
import org.xdat.gui.tables.DataSheetTableColumnModel;

/**
 * Representation of Parallel sets chart.
 *
 * This class extends ParallelChart and stores data for parallel sets chart
 * such as parallel sets axes, filters and all settings for chart. This class
 * also provides methods for building of axes, categories and sets, changing
 * order of axes, creating and changing filters, generating colors and other
 * functionality of parallel sets chart.
 *
 * @author Martin Janda
 */
public class ParallelSetsChart extends ParallelChart {

    /** Number of categories on axes for automatic splitting */
    public static final int NUMBER_OF_CATEGORIES = 3;

    /** Constant for bundled layout of chart */
    public static final int BUNDLED_LAYOUT = 0;
    /** Constant for color bundled layout of chart */
    public static final int BUNDLED_COLOR_LAYOUT = 1;
    /** Constant for tree layout of chart */
    public static final int TREE_LAYOUT = 2;

    /** Constant for coordinate axes without scaling */
    public static final int COORD_AXIS_NO_SCALING = 0;
    /** Constant for coordinate axes with scaling of cetegories */
    public static final int COORD_AXIS_SCALING_CATEGORIES = 1;
    /** Constant for coordinate axes with scaling of axes */
    public static final int COORD_AXIS_SCALING_AXIS = 2;

    /** Color of parallelograms without filter */
    public static final Color NO_FILTER_COLOR = new Color(130, 90, 90, 40);
    
    public static final int LEFT_MARGIN = 50;
    public static final int RIGHT_MARGIN = 150;

    /** Flag to enable debug message printing for this class */
    static final boolean printLog = true;

    /** Layout of parallel sets chart (bundled or tree layout) */
    private int setsLayoutMode = BUNDLED_LAYOUT;

    /** Mode of scaling coordinate axes */
    private int coordAxisScaleMode = COORD_AXIS_NO_SCALING;

    /** Flag to enable parallelograms inside axes */
    boolean insideParallelogramsOn = false;

    /** All axes of this chart */
    protected List<ParallelSetsAxis> parallelSetsAxes = new ArrayList<>();
    protected List<ParallelSetsAxis> displayAxes = new ArrayList<>();
    
    /** Axis created from active filters */
    protected ParallelSetsFilterAxis filterAxis;
    
    /** Axis according to which parallelograms are separated and colored */
    private ParallelSetsAxis separatingAxis;
    
    /** All filters of this chart */
    List<ParallelSetsFilter> filters;

    /** Filter selected by user on filter axis */
    ParallelSetsFilter selectedFilter;

    /** Filters are only on one axis */
    private boolean oneAxis = true;
    
    //private Set<Color> setsColors;
    private HashMap<Color, Integer> setsColors;

    /** Sort button at axes is enabled */
    private boolean sortButtonOn = true;
    
    /** Flag indicating need to build axes */
    private boolean buildAxesRequired = true;
    /** Flag indicating need to update axes */
    private boolean updateAxesRequired = true;
    /** Flag indicating need to update categories */
    private boolean updateCategoriesRequired = true;
    /** Flag indicating need to build design sets (parallelograms) */
    private boolean buildDesignSetsRequired = true;
    /** Flag indicating need to update design sets (parallelograms) */
    private boolean updateDesignSetsRequired = true;
    /** Flag indicating need to update designs */
    private boolean updateDesignsRequired = true;
    /** Flag indicating need to build filter axis */
    private boolean buildAxisFromFiltersRequired = true;

    private int standardAxisBoxWidth = ParallelSetsAxis.DEFAULT_BOX_WIDTH;
    private int coordinateAxisBoxWidth = ParallelSetsAxis.BOX_WIDTH_WITH_COORDINATES;

    private int allAxesWidth;

    /**
     * Instantiates a new parallel sets chart.
     * @param dataSheet
     * @param columnModel
     * @param progressMonitor
     * @param id
     */
    public ParallelSetsChart(DataSheet dataSheet, DataSheetTableColumnModel columnModel, ProgressMonitor progressMonitor, int id) {
        super(dataSheet, progressMonitor, id);

        setsColors = new HashMap<>();
        
        log("constructor: Base settings read. Creating axes...");
        progressMonitor.setMaximum(dataSheet.getParameterCount() - 1);
        progressMonitor.setNote("Building axes...");
        for (int i = 0; i < dataSheet.getParameterCount() && !progressMonitor.isCanceled(); i++) {
                Parameter p = dataSheet.getParameter(i);
                ParallelSetsAxis newAxis = new ParallelSetsAxis(dataSheet, this, p);
                parallelSetsAxes.add(newAxis);
                progressMonitor.setProgress(i);
        }
        progressMonitor.setNote("Building axis clones...");
        for (int i = 0; i < dataSheet.getParameterCount() && !progressMonitor.isCanceled(); i++) {
                Parameter p = dataSheet.getParameter(i);
                if(!p.getChildren().isEmpty()) {
                    ParallelSetsAxis newAxis = new ParallelSetsAxis(dataSheet, this, p);
                    p.setAxisClone(newAxis);
                }
                progressMonitor.setProgress(i);
        }
        
        buildDisplayAxes();
        
        List<ParallelSetsAxis> allAxes = getAllAxes();
        progressMonitor.setMaximum(allAxes.size() - 1);
        progressMonitor.setNote("Building categories...");
        for (int i = 0; i < allAxes.size() && !progressMonitor.isCanceled(); i++) {
                buildAxisCategories(allAxes.get(i));
                progressMonitor.setProgress(i);
        }

        sortAxesCategories();
        
        for(ParallelSetsAxis axis : getDisplayAxes()) {
            if(axis.isActive()) {
                if(axis.getCategories().size() > 20) continue;
                separatingAxis = axis;
                separatingAxis.setSelected(true);
                break;
            }
        }
        // init filters
        filters = new LinkedList<>();
        initFilters();
        updateFilterOnDesigns();

        buildDesignSets();
        createAxisFromFilters();

        UserPreferences userPreferences = UserPreferences.getInstance();
        allAxesWidth = userPreferences.getParallelCoordinatesAxisWidth();
    }

    public int getAllAxesWidth() {
        return allAxesWidth;
    }

    public void setAllAxesWidth(int allAxesWidth) {
        this.allAxesWidth = allAxesWidth;
    }

    public int getStandardAxisBoxWidth() {
        return standardAxisBoxWidth;
    }

    public void setStandardAxisBoxWidth(int standardAxisBoxWidth) {
        this.standardAxisBoxWidth = standardAxisBoxWidth;
    }

    public int getCoordinateAxisBoxWidth() {
        return coordinateAxisBoxWidth;
    }

    public void setCoordinateAxisBoxWidth(int coordinateAxisBoxWidth) {
        this.coordinateAxisBoxWidth = coordinateAxisBoxWidth;
    }

    public void setBuildAxisFromFiltersRequired(boolean buildAxisFromFiltersRequired) {
        this.buildAxisFromFiltersRequired = buildAxisFromFiltersRequired;
    }

    public void updateFilterOnDesigns(){
        List<ParallelSetsFilter> reversedFilters = new ArrayList<>(filters);
        Collections.reverse(reversedFilters);

        for (Design design : dataSheet.getDesigns()) {
            design.setFilter(null);
            for (ParallelSetsFilter filter : reversedFilters) {
                if(filter.test(design)){
                    design.setFilter(filter);
                }
            }
        }
    }
    
    @Override
    public void hideAxis(Axis axis){
        axis.setActive(false);
        buildAxesRequired = true;
        updateAxesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    @Override
    public void showAxis(Axis axis){
        axis.setActive(true);
        buildAxesRequired = true;
        updateAxesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    @Override
    public void moveAxisLeft(Axis axis, DataSheetTableColumnModel cm){
        super.moveAxisLeft(axis, cm);
        updateAxesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    @Override
    public void moveAxisRight(Axis axis, DataSheetTableColumnModel cm){
        super.moveAxisRight(axis, cm);
        updateAxesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    public void splitCategory(AxisCategory category, double value){
        if(!category.getCondition().getClass().equals(AxisCategoryRangeCondition.class)) return;
        
        ParallelSetsAxis axis = category.getAxis();
        AxisCategoryRangeCondition condition = (AxisCategoryRangeCondition) category.getCondition();
        AxisCategoryRangeCondition c1 = new AxisCategoryRangeCondition(condition.getMin(), value, axis.getParameter());
        AxisCategoryRangeCondition c2 = new AxisCategoryRangeCondition(value, condition.getMax(), condition.isClosedEndRange(), axis.getParameter());

        ArrayList<Design> designs1 = new ArrayList<>();
        ArrayList<Design> designs2 = new ArrayList<>();
        for(Design d : category.getDesigns()) {
            if(c1.test(d)) {
                designs1.add(d);
            }
            else {
                designs2.add(d);
            }
        }

        double heightRatio1 = ((double)designs1.size() / (double)category.getDesigns().size()) * category.getHeightRatio();
        double heightRatio2 = ((double)designs2.size() / (double)category.getDesigns().size()) * category.getHeightRatio();

        AxisCategory cat1 = new AxisCategory(c1, heightRatio1, designs1, axis);
        AxisCategory cat2 = new AxisCategory(c2, heightRatio2, designs2, axis);
        int index = axis.getCategories().indexOf(category);
        axis.getCategories().remove(index);
        axis.getCategories().add(index, cat1);
        axis.getCategories().add(index, cat2);
        
        updateAxesRequired = true;
        updateDesignsRequired = true;
        updateCategoriesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    public void mergeCategories(List<AxisCategory> selectedCategories){
        double minValue = Double.MAX_VALUE;
        double maxValue = Double.MIN_VALUE;
        AxisCategoryRangeCondition maxCondition = null;
        double heightRatio = 0;
        AxisCategory cat = selectedCategories.get(0);
        ParallelSetsAxis axis = cat.getAxis();
        List<Design> designs = new ArrayList<>();
        int index = axis.getCategories().indexOf(cat);
        for (AxisCategory category : selectedCategories) {
            AxisCategoryRangeCondition c = (AxisCategoryRangeCondition) category.getCondition();
            if(c.getMin() < minValue) minValue = c.getMin();
            if(c.getMax() > maxValue) {
                maxValue = c.getMax();
                maxCondition = c;
            }
            heightRatio += category.getHeightRatio();
            designs.addAll(category.getDesigns());

            int i = axis.getCategories().indexOf(category);
            if(i < index) index = i;
        }
        AxisCategoryRangeCondition c = new AxisCategoryRangeCondition(minValue, maxValue, maxCondition.isClosedEndRange(), axis.getParameter());
        AxisCategory newCategory = new AxisCategory(c, heightRatio, designs, axis);

        axis.getCategories().removeAll(selectedCategories);
        axis.getCategories().add(index, newCategory);
        
        updateAxesRequired = true;
        updateDesignsRequired = true;
        updateCategoriesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    @Override
    public void setFrameSize(Dimension size){
        super.setFrameSize(size);
        updateAxesRequired = true;
        updateDesignSetsRequired = true;
    }
    
    public void showCoordinatesOnAxis(ParallelSetsAxis axis, boolean showCoord){
        axis.setCoordinatesAxis(showCoord);
        updateAxesRequired = true;
        updateDesignSetsRequired = true;
    }
    
    public void invertAxis(ParallelSetsAxis axis){
        if(axis != null){
            axis.setInverted(!axis.isInverted());
        }
        
        updateAxesRequired = true;
        updateCategoriesRequired = true;
        updateDesignSetsRequired = true;
    }
    
    public boolean isSortButtonOn() {
        return sortButtonOn;
    }

    public void setSortButtonOn(boolean sortButtonOn) {
        this.sortButtonOn = sortButtonOn;
    }

    public int getCoordAxisScaleMode() {
        return coordAxisScaleMode;
    }

    public void setCoordAxisScaleMode(int coordAxisScaleMode) {
        this.coordAxisScaleMode = coordAxisScaleMode;
        
        for (ParallelSetsAxis axis : getAllAxes()) {
            if(axis.isCoordinatesAxis()){
                axis.updateCategoriesHeightRatio();
            }
        }
        
        updateAxesRequired = true;
        updateDesignSetsRequired = true;
    }
    
    public void changeFilterColor(ParallelSetsFilter filter, Color color){
        filter.setColor(color);
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    public void removeFilter(ParallelSetsFilter filter){
        if(filter == null) return;
        filters.remove(filter);
        setSelectedFilter(null);

        buildAxisFromFiltersRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    public boolean isInsideParallelogramsOn() {
        return insideParallelogramsOn && getLayoutMode() == BUNDLED_LAYOUT;
    }

    public void setInsideParallelogramsOn(boolean insideSetsOn) {
        this.insideParallelogramsOn = insideSetsOn;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    public ParallelSetsFilter getSelectedFilter() {
        return selectedFilter;
    }

    public void setSelectedFilter(ParallelSetsFilter selectedFilter) {
        this.selectedFilter = selectedFilter;
        
        clearSelectedCategories();
        if(selectedFilter != null) {
            for(AxisCategory category : selectedFilter.getCategories()){
                category.setSelected(true);
            }
        }
        
    }

    public boolean getOneAxis() {
        return oneAxis;
    }

    public void setOneAxis(boolean oneAxis) {
        this.oneAxis = oneAxis;
    }

    public void buildFiltersForSelectedAxis(){
        for(AxisCategory selectedCategory : separatingAxis.getCategories()){
            ParallelSetsFilter mAFilter = new ParallelSetsFilter(filters.size());
            mAFilter.addCategory(selectedCategory);
            filters.add(mAFilter);
            selectedFilter = mAFilter;
        }
        
        generateFiltersColors();
    }
    
    
    private void initFilters(){
        clearSelectedCategories();
        selectedFilter = null;
        filters.clear();
        for(AxisCategory category : separatingAxis.getCategories()){
            category.setSelected(true);
        }
        buildFiltersForSelectedAxis();
    }

    public List<ParallelSetsFilter> getFilters() {
        return filters;
    }

    public void setFilters(List<ParallelSetsFilter> filters) {
        this.filters = filters;
    }

    public void setLayoutMode(int mode){
        setsLayoutMode = mode;
        oneAxis = true;
        initFilters();

        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    public int getLayoutMode() {
        return setsLayoutMode;
    }

    public void clearSelectedCategories(){
        for (ParallelSetsAxis axis : getDisplayAxes()) {
            for (AxisCategory category : axis.getCategories()) {
                category.setSelected(false);
            }
        }
    }
    
    public void switchSelectedCategory(AxisCategory category){
        if(category.isSelected()){
            removeSelectedCategory(category);
        }else{
            addSelectedCategory(category);
        }
    }
    
    public void addSelectedCategory(AxisCategory category){
        category.setSelected(true);
        if(selectedFilter == null){
            selectedFilter = new ParallelSetsFilter(filters.size());
            filters.add(selectedFilter);
            generateFiltersColors();
        }
        selectedFilter.addCategory(category);

        buildAxisFromFiltersRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    public void removeSelectedCategory(AxisCategory category){
        category.setSelected(false);
        selectedFilter.removeCategory(category);
        if(selectedFilter.size() == 0) {
            filters.remove(selectedFilter);
            selectedFilter = null;
        }

        buildAxisFromFiltersRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    public void createCustomAxisByDialog(List<AxisCategory> categories) {
        AxisCategoryOrCondition previousConditions = new AxisCategoryOrCondition();
        for(AxisCategory category : categories){
            if(previousConditions.size() > 0) {
                AxisCategoryAndCondition andCondition = new AxisCategoryAndCondition(category.getCondition(), new AxisCategoryNotCondition(previousConditions.clone()));
                andCondition.setLabel(category.getCondition().getLabel());
                category.setCondition(andCondition);
            }
            previousConditions.addCondition(category.getCondition());
        }

        DataSheet dataSheet = getDataSheet();
        ParallelSetsAxisWithoutParameter newAxis = new ParallelSetsAxisWithoutParameter(dataSheet, this, "Custom Axis");
        parallelSetsAxes.add(newAxis);
        buildDisplayAxes();

        List<Design> noCategoryDesigns = new ArrayList<>();
        AxisCategoryOrCondition anyCategoryCondition = new AxisCategoryOrCondition();
        for(int i = 0; i < dataSheet.getDesignCount(); i++) {
            Design d = dataSheet.getDesign(i);
            if(!d.isActive()) continue;
            boolean inCategory = false;

            for(AxisCategory cat : categories) {
                if(cat.getCondition().test(d)) {
                    inCategory = true;
                    cat.getDesigns().add(d);
                    break;
                }
            }
            if(!inCategory) {
                noCategoryDesigns.add(d);
            }
        }


        for(AxisCategory category : categories) {
            List<Design> designsInCat = category.getDesigns();
            int designsInCatSize = 0;
            if(designsInCat != null) {
                designsInCatSize = designsInCat.size();
            }
            double designsCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
            double height = (double) designsInCatSize / designsCount;
            category.setHeightRatio(height);
            category.setAxis(newAxis);
            anyCategoryCondition.addCondition(category.getCondition());
        }

        int designsInCatSize = noCategoryDesigns.size();
        double designsCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
        double height = (double) designsInCatSize / designsCount;
        AxisCategory others = new AxisCategory(new AxisCategoryNotCondition(anyCategoryCondition), height, noCategoryDesigns, newAxis);
        categories.add(others);

        newAxis.setCategories(categories);
        newAxis.update(6);

        updateAxesRequired = true;
        updateCategoriesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    public void createAxisFromFilters() {
        List<AxisCategory> categories = new ArrayList<>();
        filterAxis = new ParallelSetsFilterAxis(this);
        DataSheet dataSheet = getDataSheet();
        
        HashMap<ParallelSetsFilter, List<Design>> categoryMap = new HashMap<>();
        List<Design> noFilterDesigns = new ArrayList<>();
        AxisCategoryOrCondition anyFilterCondition = new AxisCategoryOrCondition();
        for(int i = 0; i < dataSheet.getDesignCount(); i++) {
            Design d = dataSheet.getDesign(i);
            if(!d.isActive()) continue;
            boolean inFilter = false;

            for(ParallelSetsFilter filter : getFilters()) {
                if(filter.test(d)) {
                    inFilter = true;
                    List<Design> designsInCat = categoryMap.get(filter);
                    if(designsInCat == null) {
                        designsInCat = new ArrayList<>();
                        categoryMap.put(filter, designsInCat);
                    }
                    designsInCat.add(d);
                    break;
                }
            }
            if(!inFilter) {
                noFilterDesigns.add(d);
            }
            buildAxisFromFiltersRequired = false;
        }
        
        double heightSum = 0;
        for(ParallelSetsFilter filter : getFilters()) {
            List<Design> designsInCat = categoryMap.get(filter);
            int designsInCatSize = 0;
            if(designsInCat != null) {
                designsInCatSize = designsInCat.size();
            }
            double designsCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
            double height = (double) designsInCatSize / designsCount;
            heightSum += height;
            AxisCategory category = new FilterAxisCategory(filter.getCondition(), filter, height, designsInCat, filterAxis);
            anyFilterCondition.addCondition(filter.getCondition());
            categories.add(category);
        }

        int designsInCatSize = noFilterDesigns.size();
        double designsCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
        double height = (double) designsInCatSize / designsCount;
        //heightSum += height;
        filterAxis.setNoFilterCategory(
                new FilterAxisCategory(new AxisCategoryNotCondition(anyFilterCondition), null, height, noFilterDesigns, filterAxis)
        );

        filterAxis.setCategories(categories);
        double dCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
        filterAxis.setNoFilterDesigns((double) noFilterDesigns.size() / dCount);
        filterAxis.update(6);
    }

    public void setHighlightedCategory(AxisCategory category) {
        
        int alphaLow = 60;
        int alphaHigh = 180;
        
        selectedFilter = null;
        filters.clear();
        clearSelectedCategories();
        category.setSelected(true);
            
        for (AxisCategory cat : separatingAxis.getCategories()) {
            if(cat.isSelected()){
                cat.setAlpha(alphaHigh);
            }else{
                cat.setAlpha(alphaLow);
            }
            for (DesignSet inputSet : cat.getInputSets()) {
                inputSet.setParentCategory(cat);
            }
            for (DesignSet outputSet : cat.getOutputSets()) {
                outputSet.setParentCategory(cat);
            }
        }
    }

    public ParallelSetsAxis getSeparatingAxis() {
        return separatingAxis;
    }

    public void setSeparatingAxis(ParallelSetsAxis separatingAxis) {
        this.separatingAxis.setSelected(false);
        this.separatingAxis = separatingAxis;
        this.separatingAxis.setSelected(true);

        clearSelectedCategories();
        initFilters(); 

        buildAxisFromFiltersRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    private void buildDisplayAxes() {
        switch(getLayoutMode()) {
            case TREE_LAYOUT:
                displayAxes = getTreeAxes();
                break;
            case BUNDLED_LAYOUT:
            case BUNDLED_COLOR_LAYOUT:
            default:
                displayAxes = getCategoryAxes();
                break;
        }
        buildAxesRequired = false;
    }
    
    private void buildAxesCategories(){
        for(ParallelSetsAxis axis : getAllAxes()) {
            buildAxisCategories(axis);
        }
    }
    
    private void sortAxesCategories() {
        for(ParallelSetsAxis axis : getAllAxes()) {
            axis.sortCategories();
        }
    }
    
    private void buildAxisCategories(ParallelSetsAxis axis) {
        List<AxisCategory> categories = new ArrayList<>();
        DataSheet dataSheet = axis.getDataSheet();
        Parameter parameter = axis.getParameter();
        if(parameter == null) return;
        if(parameter.isNumeric()) {
            double max = dataSheet.getMaxValueOf(parameter);
            double min = dataSheet.getMinValueOf(parameter);
            double step = (max - min) / (double) NUMBER_OF_CATEGORIES;
            double heightSum = 0;
            int numberOfCat = NUMBER_OF_CATEGORIES;
            
            if(parameter.isOneNonNumeric()){
                numberOfCat++;
            }
            
            AxisCategoryCondition[] cats = new AxisCategoryCondition[numberOfCat];
            ArrayList<Design>[] designs = new ArrayList[numberOfCat];

            double from = min;
            double to = from + step;
            for(int j = 0; j < NUMBER_OF_CATEGORIES; j++) {
                AxisCategoryCondition c = null;
                if(j+1 == NUMBER_OF_CATEGORIES) {
                    to = max;
                    c = new AxisCategoryRangeCondition(from, to, true, parameter);
                }
                else {
                    c = new AxisCategoryRangeCondition(from, to, parameter);
                }

                cats[j] = c;
                designs[j] = new ArrayList<>();
                from = to;
                to += step;
            }
            
            if(parameter.isOneNonNumeric()){
                cats[numberOfCat-1] = new AxisCategoryStringCondition(parameter.getOtherValueString(), parameter);
                designs[numberOfCat-1] = new ArrayList<>();
            }

            for(int i = 0; i < dataSheet.getDesignCount(); i++) {
                Design d = dataSheet.getDesign(i);
                if(!d.isActive()) continue;
                for(int j = 0; j < cats.length; j++) {
                    AxisCategoryCondition c = cats[j];
                    if(c.test(d)){
                        List<Design> designsInCat = designs[j];
                        designsInCat.add(d);
                    }
                }
            }

            // Create the categories
            for(int j = 0; j < cats.length; j++) {
                AxisCategoryCondition c = cats[j];
                List<Design> designsInCat = designs[j];
                if(designsInCat.isEmpty()) continue;
                double designsCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
                double height;
                if(axis.isCoordinatesAxis() && coordAxisScaleMode == COORD_AXIS_SCALING_CATEGORIES){
                    AxisCategoryRangeCondition rangeCondition = (AxisCategoryRangeCondition) c;
                    height = (rangeCondition.getMax() - rangeCondition.getMin()) / axis.getRange();
                }else{
                    height = (double) designsInCat.size() / designsCount;
                }
                heightSum += height;
                categories.add(new AxisCategory(c, height, designsInCat, axis));
            }


            if(heightSum != 1.0) System.out.println("Wrong height sum for axis " + axis.getName());
            axis.setCategories(categories);
        }
        else {
            HashMap<String, List<Design>> categoryMap = new HashMap<>();
            for(int i = 0; i < dataSheet.getDesignCount(); i++) {
                Design d = dataSheet.getDesign(i);
                if(!d.isActive()) continue;
                String value = d.getStringValue(parameter);
                List<Design> designsInCat = categoryMap.get(value);
                if(designsInCat == null) {
                    designsInCat = new ArrayList<>();
                    categoryMap.put(value, designsInCat);
                }
                designsInCat.add(d);
            }
            double heightSum = 0;
            for(String catName : categoryMap.keySet()) {
                List<Design> designsInCat = categoryMap.get(catName);
                double designsCount = (double) (dataSheet.getDesignCount() - dataSheet.getInactiveDesigns());
                double height = (double) designsInCat.size() / designsCount;
                heightSum += height;
                categories.add(new AxisCategory(new AxisCategoryStringCondition(catName, parameter), height, designsInCat, axis));
            }
            if(heightSum != 1.0) System.out.println("Wrong height sum for axis " + axis.getName());
            axis.setCategories(categories);
        }
    }

    /**
     * Build design sets for tree layout
     * @param fromAxis
     * @param toAxis
     * @param reversed 
     */
    private void buildDesignSetsBetweenAxes(ParallelSetsAxis fromAxis, ParallelSetsAxis toAxis, boolean reversed){
        boolean rootPass = false;
        if(fromAxis == separatingAxis) rootPass = true;
        
        // clear pervious sets
        for(AxisCategory cat : toAxis.getCategories()){
            cat.getInputSets().clear();
            cat.getOutputSets().clear();
        }
        int catFromIndex = 0;
        for(AxisCategory catFrom : fromAxis.getCategories()){ // all from categories
            int catToIndex = 0;
            for(AxisCategory catTo : toAxis.getCategories()){ // all to categories
                List<Design> outputDesigns = catTo.getDesigns();
                List<DesignSet> inputSets;
                if(reversed){
                    inputSets = catFrom.getOutputSets();
                }else{
                    inputSets = catFrom.getInputSets();
                }
                if(rootPass){ // if axis is root
                    inputSets = new ArrayList<>();
                    inputSets.add(new DesignSet(catFromIndex, -1, catFrom, null, catFrom.getDesigns(), catFrom));
                }
                for(DesignSet inputSet : inputSets){ // all input set
                    List<Design> designs = new ArrayList<>();
                    for(Design d : inputSet.getDesigns()){
                        if(outputDesigns.contains(d)){
                            designs.add(d);
                        }
                    }
                    if(designs.isEmpty()) continue;

                    List<DesignSet> fromSets;
                    List<DesignSet> toSets;
                    if(reversed){
                        fromSets = catFrom.getInputSets();
                        toSets = catTo.getOutputSets();
                    }else{
                        fromSets = catFrom.getOutputSets();
                        toSets = catTo.getInputSets();
                    }

                    int inputIndex = catFromIndex;
                    int outputIndex = catToIndex;
                    AxisCategory inputCategory = catFrom;
                    AxisCategory outputCategory = catTo;
                    if(reversed){
                        inputIndex = catToIndex;
                        outputIndex = catFromIndex;
                        inputCategory = catTo;
                        outputCategory = catFrom;
                    }
                    
                    DesignSet set;
                    if(rootPass){
                        set = new DesignSet(inputIndex, outputIndex, inputCategory, outputCategory, designs, catFrom);
                    }else{
                        set = new DesignSet(inputIndex, outputIndex, inputCategory, outputCategory, designs, inputSet);
                    }
                    
                    fromSets.add(set);
                    toSets.add(set);
                    inputSet.getChildSets().add(set);
                }
                catToIndex++;
            }
            catFromIndex++;
        }
    }
    
    public void buildDesignSets(){
        if(setsLayoutMode == TREE_LAYOUT){
            buildTreeDesignSets();
        }else{
            buildBundledDesignSets();
        }
        buildDesignSetsRequired = false;
    }
    
    public void buildTreeDesignSets(){
        log("buildDesignSetsTree");
        int startIndex = displayAxes.indexOf(separatingAxis);
        generateCategoriesColors();
        
        // clear root
        for(AxisCategory cat : separatingAxis.getCategories()){
            cat.getInputSets().clear();
            cat.getOutputSets().clear();
        }
        
        int nextIndex = startIndex + 1;
        int lastIndex = startIndex - 1;
        
        // forward pass
        if(nextIndex < displayAxes.size()){
            ParallelSetsAxis lastAxis = separatingAxis;
            for(int axisI = nextIndex; axisI < getAxisCount(); axisI++){
                ParallelSetsAxis currentAxis = getAxis(axisI);
                if(!currentAxis.isActive()) continue;
                buildDesignSetsBetweenAxes(lastAxis, currentAxis, false);
                lastAxis = currentAxis;
            }
        }
        
        // back pass
        if(lastIndex >= 0){
            ParallelSetsAxis lastAxis = separatingAxis;
            for(int axisI = lastIndex; axisI >= 0; axisI--){
                ParallelSetsAxis currentAxis = getAxis(axisI);
                if(!currentAxis.isActive()) continue;
                buildDesignSetsBetweenAxes(lastAxis, currentAxis, true);
                lastAxis = currentAxis;
            }
        }
        
    }

    public void buildBundledDesignSets(){

        ParallelSetsAxis lastAxis = null;
        for(ParallelSetsAxis currentAxis : getDisplayAxes()) { // all axes (pairs)
            // clear pervious sets
            for(AxisCategory cat : currentAxis.getCategories()){
                cat.getInputSets().clear();
                cat.getOutputSets().clear();
                cat.getInsideSets().clear();
            }
            if(lastAxis == null){ // skip first axis, save as last
                lastAxis = currentAxis;
                continue;
            }

            // Design sets between axis pairs
            for(AxisCategory catFrom : lastAxis.getCategories()){ // all from categories
                HashMap<AxisCategory, HashMap<Color, List<Design>>> designsMap = new HashMap<>();
                List<Design> inputDesigns = catFrom.getDesigns();
                for(Design d : inputDesigns) { // all input designs
                    for(AxisCategory catTo : currentAxis.getCategories()) { // all to categories
                        if(!catTo.getCondition().test(d)) continue;
                        Color c = NO_FILTER_COLOR;
                        int filterId = Integer.MAX_VALUE;
                        if(d.getFilter() != null){
                            c = d.getFilter().getColor();
                            filterId = d.getFilter().getFilterID();
                        }
                        HashMap<Color, List<Design>> designsByColor = designsMap.get(catTo);
                        if(designsByColor == null) {
                            designsByColor = new HashMap<>();
                            designsMap.put(catTo, designsByColor);
                        }
                        List<Design> designs = designsByColor.get(c);
                        if(designs == null) {
                            designs = new ArrayList<>();
                            designsByColor.put(c, designs);
                        }

                        d.setColor(c);
                        d.setFilterId(filterId);
                        designs.add(d);
                        break; // add design only to one category
                    }
                }

                for(AxisCategory catTo : designsMap.keySet()) {
                    HashMap<Color, List<Design>> designsByColor = designsMap.get(catTo);
                    int setIndex = 0;
                    for(Color color : designsByColor.keySet()){
                        List<Design> designs = designsByColor.get(color);
                        if(designs.isEmpty()) continue;

                        ParallelSetsFilter filter = designs.get(0).getFilter();
                        DesignSet set = new DesignSet(lastAxis.getCategories().indexOf(catFrom), currentAxis.getCategories().indexOf(catTo), catFrom, catTo, designs, filter);
                        catFrom.getOutputSets().add(set);
                        catTo.getInputSets().add(set);

                        // build INSIDE sets in lastAxis
                        if(color.equals(NO_FILTER_COLOR) || !isInsideParallelogramsOn()) continue;
                        int inputSetIndex = 0;
                        for(DesignSet inputSet : catFrom.getInputSets()){
                            if(inputSet.getFilter() == null) continue;
                            if(inputSet.getFilter().getFilterID() != filter.getFilterID()) continue;

                            List<Design> inAxisSetDesigns = new ArrayList<>();
                            for (Design design : set.getDesigns()) {
                                if(inputSet.getDesigns().contains(design)) inAxisSetDesigns.add(design);
                            }

                            if(!inAxisSetDesigns.isEmpty()) {
                                InsideDesignSet insideSet = new InsideDesignSet(catFrom, inputSetIndex, setIndex, inputSet, set, inAxisSetDesigns, color);
                                catFrom.getInsideSets().add(insideSet);
                                inputSet.getOutputInsideSets().add(insideSet);
                                set.getInputInsideSets().add(insideSet);
                            }
                            inputSetIndex++;
                        }
                        setIndex++;
                    }
                }
            }
            lastAxis = currentAxis;
        }
    }
    
    public void generateFiltersColors(){
        float step = 1.0f / filters.size();
        for(int i = 0; i < filters.size(); i++){
            int alpha = 128;

            Color c = Color.getHSBColor(i * step, 0.8f, 1f);
            Color transpColor = new Color(c.getRed(), c.getGreen(), c.getBlue(), alpha);
            filters.get(i).setColor(transpColor);
        }
    }
    
    public Color[] generateColorsGeneral(int number){
        Color[] colors = new Color[number];
        float step = 1.0f / number;
        for(int i = 0; i < number; i++){
            int alpha = 100;

            Color c = Color.getHSBColor(i * step, 1, 1);
            Color transpColor = new Color(c.getRed(), c.getGreen(), c.getBlue(), alpha);
            colors[i] = transpColor;
        }
        return colors;
    }
    
    public void generateCategoriesColors(){
        float step = 1.0f / separatingAxis.getNumberOfActiveCategories();
        for(int i = 0; i < separatingAxis.getNumberOfActiveCategories(); i++){
            AxisCategory category = separatingAxis.getCategoryFromActive(i);
            
            int alpha = 100;
            if(!oneAxis) alpha = 60;

            Color c = Color.getHSBColor(i * step, 1, 1);
            Color transpColor = new Color(c.getRed(), c.getGreen(), c.getBlue(), alpha);
            category.setColor(transpColor);
            if(!oneAxis && category.isSelected()) category.setAlpha(180);
        }
    }
    
    public List<ParallelSetsAxis> getAllAxes() {
        LinkedList<ParallelSetsAxis> axes = new LinkedList<>();
        for(ParallelSetsAxis axis : parallelSetsAxes) {
            Parameter p = axis.getParameter();
            if(p == null){
                axes.add(axis);
                continue;
            }
            if(p.isAggregated()) continue;
            
            axes.add(axis); 
            for(Parameter child : axis.getParameter().getChildren()) {
                ParallelSetsAxis childAxis = (ParallelSetsAxis) child.getAxis();
                axes.add(childAxis);
            }
            ParallelSetsAxis clone = (ParallelSetsAxis) p.getAxisClone();
            if(clone != null) {
                axes.add(clone);
            }
        }
        return axes;
    }
    
    public List<ParallelSetsAxis> getDisplayAxes() {
        return displayAxes;
    }
    
    public List<ParallelSetsAxis> getCategoryAxes() {
        LinkedList<ParallelSetsAxis> axes = new LinkedList<>();
        for(ParallelSetsAxis axis : parallelSetsAxes) {
            if(!axis.isActive()) continue;
            Parameter p = axis.getParameter();
            if(p != null && p.isAggregated()) continue;
            
            axes.add(axis); 
            if(axis.isExpanded()) {
                int children = 0;
                for(Parameter child : axis.getParameter().getChildren()) {
                    ParallelSetsAxis childAxis = (ParallelSetsAxis) child.getAxis();
                    if(!childAxis.isActive()) continue;
                    axes.add(childAxis);
                    children++;
                }
                ParallelSetsAxis clone = (ParallelSetsAxis) p.getAxisClone();
                if(clone != null && children > 0) {
                    axes.add(clone);
                }
            }
        }
        return axes;
    }
    
    public List<ParallelSetsAxis> getTreeAxes() {
        LinkedList<ParallelSetsAxis> axes = new LinkedList<>();
        for(ParallelSetsAxis axis : parallelSetsAxes) {
            if(!axis.isActive()) continue;
            Parameter p = axis.getParameter();
            if(p != null && p.isAggregated()) continue;
            axes.add(axis); 
        }
        return axes;
    }
    
    @Override
    public ParallelSetsAxis getAxis(int index) {
        return displayAxes.get(index);
    }
    
    public int getAxisCount() {
        return displayAxes.size();
    }
    
    public void addAxis(ParallelSetsAxis axis) {
        parallelSetsAxes.add(axis);
        
        buildDisplayAxes();
        buildDesignSets();
    }
    
    public void addAxis(int index, ParallelSetsAxis axis) {
        parallelSetsAxes.add(index, axis);
        
        buildDisplayAxes();
        buildDesignSets();
    }
    
    public void removeAxis(int index) {
        parallelSetsAxes.remove(index);
        
        buildDisplayAxes();
        buildDesignSets();
    }

    public void moveAxis(int oldIndex, int newIndex) {
        ParallelSetsAxis axis = parallelSetsAxes.remove(oldIndex);
        parallelSetsAxes.add(newIndex, axis);

        // rebuild axes and sets
        buildAxesRequired = true;
        updateAxesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    public void moveAxisAndParameter(int oldIndex, int newIndex, int oldPIndex, int newPIndex, DataSheetTableColumnModel columnModel) {
	    log("moveAxis called with arguments " + oldIndex + " and " + newIndex);

	    if(parallelSetsAxes.get(oldIndex).getParameter() != null){
	        columnModel.moveColumnWithoutAxis(oldPIndex+1, newPIndex+1);
        }

	    moveAxis(oldIndex, newIndex);
    }
    
    public void moveCategory(AxisCategory category, int newIndex){
        category.getAxis().moveCategory(category, newIndex);

        updateAxesRequired = true;
        updateCategoriesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }
    
    @Override
    public int getWidth() {
        int width = 0;
        if(!displayAxes.isEmpty()) {
            ParallelSetsAxis lastAxis = null;
            for(ParallelSetsAxis axis : displayAxes) {
                if(lastAxis != null) {
                    int xOffset = lastAxis.getBoxWidth() + (int) (lastAxis.getWidth() * 0.5) + (int) (axis.getWidth() * 0.5);
                    width += xOffset;
                }
                lastAxis = axis;
            }
            width += lastAxis.getBoxWidth();
        }
        return width + RIGHT_MARGIN;
    }
    
    @Override
    public String getTitle() {
        return "Parallel Sets Chart " + this.getID();
    }
    
    /**
     * Prints debug information to stdout when printLog is set to true.
     * 
     * @param message
     *            the message
     */
    protected void log(String message) {
            if (ParallelSetsChart.printLog && Main.isLoggingEnabled()) {
                    System.out.println(this.getClass().getName() + "." + message);
            }
    }
    
    private void updateTreeSets(DesignSet set){
        // update root
        AxisCategory catFrom = set.getInputCategory();
        AxisCategory catTo = set.getOutputCategory();

        float inputSize = catFrom.getDesigns().size();
        float outputSize = catTo.getDesigns().size();
        float inputH = (catFrom.getHeight() * set.getDesigns().size()) / inputSize;
        float outputH = (catTo.getHeight() * set.getDesigns().size()) / outputSize;
        float inputShift = catFrom.getInSetsHeight();
        float outputShift = catTo.getOutSetsHeight();
        float inputY = inputShift + catFrom.getyUpPosition();
        float outputY = outputShift + catTo.getyUpPosition();
        catFrom.setInSetsHeight(inputShift + inputH);
        catTo.setOutSetsHeight(outputShift + outputH);

        set.setInputHeight((int) inputH);
        set.setInputY((int) inputY);
        set.setOutputHeight((int) outputH);
        set.setOutputY((int) outputY);

        // update all children
        for(DesignSet childSet : set.getChildSets()){
            updateTreeSets(childSet);
        }
    }

    private void clearAllCategories(){
        for(ParallelSetsAxis axis : getDisplayAxes()){
            for(AxisCategory cat : axis.getCategories()){
                cat.setOutSetsHeight(0);
                cat.setInSetsHeight(0);
            }
        }
    }
    
    public void updateAxes() {
        // Update axes
        List<ParallelSetsAxis> displayAxes = getDisplayAxes();
        if(!displayAxes.isEmpty()) {
            ParallelSetsAxis first = displayAxes.get(0);
            int xPosition = LEFT_MARGIN;
            ParallelSetsAxis lastAxis = null;
            for(ParallelSetsAxis axis : displayAxes) {
                if(lastAxis != null) {
                    int xOffset = lastAxis.getBoxWidth() + (int) (lastAxis.getWidth() * 0.5) + (int) (axis.getWidth() * 0.5);
                    xPosition += xOffset;
                }
                axis.update(xPosition);
                lastAxis = axis;
            }
        }
        filterAxis.update(6);
        updateAxesRequired = false;
    }
        
    public void updateTreeLayout() {
        ParallelSetsAxis currentAxis = getSeparatingAxis();
        // forward pass
        for(AxisCategory catFrom : currentAxis.getCategories()){
            for(DesignSet set : catFrom.getOutputSets()){
                updateTreeSets(set);
            }
        }
        // back pass
        for(AxisCategory catFrom : currentAxis.getCategories()){
            for(DesignSet set : catFrom.getInputSets()){
                updateTreeSets(set);
            }
        }   
    }

    public void updateColorLayout() {
        ParallelSetsAxis lastAxis = null;
        for(ParallelSetsAxis currentAxis : getDisplayAxes()){ // all axis pairs
            if(!currentAxis.isActive()) continue;
            if(lastAxis == null){
                lastAxis = currentAxis;
                continue;
            }
            for(AxisCategory catFrom : lastAxis.getCategories()){ // all categories
                for(int i = 0; i <= getFilters().size(); i++){
                    Color color = NO_FILTER_COLOR;
                    if(i < getFilters().size()){
                        color = getFilters().get(i).getColor();
                    }
                    for(AxisCategory catTo : currentAxis.getCategories()){
                        for(DesignSet set : catFrom.getOutputSets()){
                            if(set.getOutputCategory() != catTo) continue;
                            //if(set.getColor() != color) continue;

                            float inputSize = catFrom.getDesigns().size();
                            float outputSize = catTo.getDesigns().size();
                            float inputH = (catFrom.getHeight() * set.getDesigns().size()) / inputSize;
                            float outputH = (catTo.getHeight() * set.getDesigns().size()) / outputSize;
                            float inputShift = catFrom.getInSetsHeight();
                            float outputShift = catTo.getOutSetsHeight();
                            float inputY = inputShift + catFrom.getyUpPosition();
                            float outputY = outputShift + catTo.getyUpPosition();
                            catFrom.setInSetsHeight(inputShift + inputH);
                            catTo.setOutSetsHeight(outputShift + outputH);

                            set.setInputHeight((int) inputH);
                            set.setInputY((int) inputY);
                            set.setOutputHeight((int) outputH);
                            set.setOutputY((int) outputY);

                            // update inside sets
                            for (InsideDesignSet insideSet : set.getInputInsideSets()) {
                                DesignSet lastSet = insideSet.getInputSet();

                                float setSize = set.getDesigns().size();
                                int insideSetSize = insideSet.getDesigns().size();
                                float height = (set.getInputHeight() * insideSetSize) / setSize;

                                inputShift = lastSet.getOutputTempHeight();
                                outputShift = set.getInputTempHeight();
                                inputY = inputShift + lastSet.getOutputY();
                                outputY = outputShift + set.getInputY();

                                set.increaseInputTempHeight(height);
                                lastSet.increaseOutputTempHeight(height);

                                insideSet.setHeight((int) height);
                                insideSet.setInputY((int) inputY);
                                insideSet.setOutputY((int) outputY);
                            }
                        }
                    }
                }
            }
            lastAxis = currentAxis;
        }
    }

    public void updateBundledLayout() {
        sortBundledDesignSets();
        for(ParallelSetsAxis axis : getDisplayAxes()) {
            for (AxisCategory catFrom : axis.getCategories()) {
                for (DesignSet set : catFrom.getOutputSets()) {
                    AxisCategory catTo = set.getOutputCategory();
                    
                    float inputSize = catFrom.getDesigns().size();
                    float inputH = catFrom.getHeight() * set.getDesigns().size() / inputSize;
                    float inputShift = catFrom.getInSetsHeight();
                    float inputY = inputShift + catFrom.getyUpPosition();

                    catFrom.setInSetsHeight(inputShift + inputH);

                    set.setInputHeight((int) inputH);
                    set.setInputY((int) inputY);

                    float outputSize = catTo.getDesigns().size();
                    float outputH = catTo.getHeight() * set.getDesigns().size() / outputSize;
                    float outputShift = catTo.getOutSetsHeight();
                    float outputY = outputShift + catTo.getyUpPosition();

                    catTo.setOutSetsHeight(outputShift + outputH);

                    set.setOutputHeight((int) outputH);
                    set.setOutputY((int) outputY);
                    
                    //update inside sets
                    for (InsideDesignSet insideSet : set.getInputInsideSets()) {
                        DesignSet lastSet = insideSet.getInputSet();

                        if(set != insideSet.getOutputSet()) 
                            System.out.println("Output set: " + set + " " + insideSet.getOutputSet());
                        if(lastSet != insideSet.getInputSet())
                            System.out.println("Input set: " + lastSet + " " + insideSet.getInputSet());

                        float setSize = set.getDesigns().size();
                        float insideSetSize = insideSet.getDesigns().size();
                        float height = set.getInputHeight() * insideSetSize / setSize;

                        inputShift = lastSet.getOutputTempHeight();
                        outputShift = set.getInputTempHeight();
                        inputY = inputShift + lastSet.getOutputY();
                        outputY = outputShift + set.getInputY();

                        set.increaseInputTempHeight(height);
                        lastSet.increaseOutputTempHeight(height);

                        insideSet.setHeight((int) height);
                        insideSet.setInputY((int) inputY);
                        insideSet.setOutputY((int) outputY);
                    }
                }
            }
        }
    }
        
    public void updateDesignSets() {
        clearAllCategories();
        // update sets of the axis
        switch(getLayoutMode()){
            case ParallelSetsChart.TREE_LAYOUT:
                updateTreeLayout();
                break;
            case ParallelSetsChart.BUNDLED_COLOR_LAYOUT:
                updateColorLayout();
                break;
            case ParallelSetsChart.BUNDLED_LAYOUT:
                updateBundledLayout();
                break;
        }
        updateDesignSetsRequired = false;
    }

    private void sortBundledDesignSets(){
        for (ParallelSetsAxis axis : parallelSetsAxes) {
            for (AxisCategory category : axis.getCategories()) {
                category.getOutputSets().sort(new Comparator<DesignSet>() {
                    @Override
                    public int compare(DesignSet set1, DesignSet set2) {
                        if(set1.getOutputCategory().getOrder() == set2.getOutputCategory().getOrder()){
                            return set1.getColorIndex() - set2.getColorIndex();
                        }else{
                            return set1.getOutputCategory().getOrder() - set2.getOutputCategory().getOrder();
                        }
                    }
                });
            }
        }
    }

    public void setBuildDesignSetsRequired(boolean buildDesignSetsRequired) {
        this.buildDesignSetsRequired = buildDesignSetsRequired;
    }

    public void update() {
        log("update (BA UC BD UA UD): "
                + buildAxesRequired + " " + updateCategoriesRequired + " " + buildDesignSetsRequired + " " + updateAxesRequired + " " + updateDesignSetsRequired);

        if(buildAxesRequired){
            buildDisplayAxes();
        }
        //buildAxesCategories();
        if(buildAxisFromFiltersRequired){
            createAxisFromFilters();
        }
        if(updateDesignsRequired){
            updateDesigns();
        }
        if(updateCategoriesRequired){
            updateCategories();
            sortAxesCategories();
        }
        if(buildDesignSetsRequired){
            updateFilterOnDesigns();
            buildDesignSets();
        }
        if(updateAxesRequired){
            updateAxes();
        }
        if(updateDesignSetsRequired){
            updateDesignSets();
        }

        sortTable();

        System.out.println("Maximum number of design sets: " + getMaxNumberOfDesignSets());
        System.out.println("Maximum number of crosings: " + getMaxNumberOfCrossings());
        System.out.println("Maximum number of inside crosings: " + getMaxNumberOfInsideCrossings());
    }

    public void incrementAxisWidth(int deltaWidth) {
        allAxesWidth = Math.max(0, allAxesWidth + deltaWidth);
    }

    public void incrementAxisBoxWidth(int deltaWidth) {
        standardAxisBoxWidth = Math.max(0, standardAxisBoxWidth + deltaWidth);
        coordinateAxisBoxWidth = Math.max(0, coordinateAxisBoxWidth + deltaWidth);
    }
    
    @Override
    public Axis getAxis(String parameterName) {
            for (Axis axis : getAllAxes()) {
                    if (axis.getParameter() != null && parameterName.equals(axis.getParameter().getName())) {
                            return axis;
                    }
            }
            throw new IllegalArgumentException("Axis " + parameterName + " not found");
    }

    private void sortTable() {
        getDataSheet().sort();
    }

    public ParallelSetsFilterAxis getFilterAxis() {
        return filterAxis;
    }
    
    public int getMaxNumberOfDesignSets() {
        boolean first = true;
        int maxNumberOfDesignSets = 0;
        int maxNumberOfInDesignSets = 0;
        for(ParallelSetsAxis axis : getDisplayAxes()) {
            if(first) {
                first = false;
                continue;
            }
            int numberOfDesignSets = 0;
            for(AxisCategory c : axis.getCategories()) {
                numberOfDesignSets += c.getInputSets().size();
                if(this.isInsideParallelogramsOn() && this.getLayoutMode() == BUNDLED_LAYOUT) {
                    int numberOfInDesignSets = c.getInsideSets().size();
                    if(maxNumberOfInDesignSets < numberOfInDesignSets) {
                        //System.out.println("Inside sets are " + numberOfInDesignSets + " for axis " + axis.getName() + " category " + c.getLabel());
                        maxNumberOfInDesignSets = numberOfInDesignSets;
                    }
                }
            }
            if(maxNumberOfDesignSets < numberOfDesignSets) {
                maxNumberOfDesignSets = numberOfDesignSets;
            }
        }
        System.out.println("Max. number of inside design sets is: " + maxNumberOfInDesignSets);
        return maxNumberOfDesignSets;
    }
    
    public int getMaxNumberOfCrossings() {
        int maxCrossings = 0;
        ParallelSetsAxis lastAxis = null;
        for(ParallelSetsAxis axis : getDisplayAxes()) {
            if(lastAxis == null) {
                lastAxis = axis;
                continue;
            }
            
            int[][] sets = new int[lastAxis.getCategories().size()][axis.getCategories().size()];
            int crossings = 0;
            int insideCrossings;
            for(AxisCategory c : axis.categories) {
                insideCrossings = 0;
                for(DesignSet set : c.getInputSets()) {
                    int outputIndex = set.getOutputIndex();
                    AxisCategory ic = set.getInputCategory();
                    int inputIndex = set.getInputIndex();
                    //System.out.print(ic.getLabel() + " index " + inputIndex + " | ");
                    //System.out.print(c.getLabel() + " index " + outputIndex);
                    
                    int setCrossings = 0;
                    for(int o = outputIndex-1; o >= 0; o--) {
                        for(int i = inputIndex+1; i < sets.length; i++) {
                            setCrossings += sets[i][o];
                        }
                    }
                    
                    //System.out.println(" crossings " + setCrossings);
                    crossings += setCrossings;
                    sets[inputIndex][outputIndex]++;
                }
                
//                if(insideSetsOn) {
//                    System.out.println("Category " + c.getLabel());
//                    int[][] insideSets = new int[c.getInputSets().size()][c.getOutputSets().size()];
//                    for(InsideDesignSet set : c.getInsideSets()) {
//                        int outputIndex = set.getOutputIndex();
//                        int inputIndex = set.getInputIndex();
//                        System.out.print("input index " + inputIndex + " | ");
//                        System.out.print("output index " + outputIndex);
//
//                        for(int o = outputIndex-1; o >= 0; o--) {
//                            for(int i = inputIndex+1; i < insideSets.length; i++) {
//                                insideCrossings += insideSets[i][o];
//                            }
//                        }
//                        insideSets[inputIndex][outputIndex]++;
//                        System.out.println(" crossings " + insideCrossings);
//                    }
//                }
                
            }
            
            //System.out.println(crossings + " crossings between axis " + lastAxis.getName() + " and " + axis.getName());
            if(crossings > maxCrossings) maxCrossings = crossings;
            lastAxis = axis;
            
        }
        return maxCrossings;
    }
    
    public int getMaxNumberOfInsideCrossings() {
        int maxCrossings = 0;
        int insideCrossings = 0;
        HashMap<DesignSet, Integer> inputSetsOrder = new HashMap<>();
        HashMap<DesignSet, Integer> outputSetsOrder = new HashMap<>();
        List<ParallelSetsAxis> axes = getDisplayAxes();
        for(ParallelSetsAxis axis : axes.subList(1, axes.size()-1)) {
            for(AxisCategory c : axis.categories) {
                if(!c.isActive()) continue;
                insideCrossings = 0;
                if(isInsideParallelogramsOn()) {
                    Comparator<DesignSet> comparator = new Comparator<DesignSet>() {
                        @Override
                        public int compare(DesignSet d1, DesignSet d2) {
                            if(d1.getInputY() < d2.getInputY()) {
                                return -1;
                            }
                            else {
                                return 1;
                            }
                        }
                    };
                    
                    List<DesignSet> inputSets = c.getInputSets();
                    Collections.sort(inputSets, comparator);
                    int index = 0;
                    for(DesignSet s : inputSets) {
                        inputSetsOrder.put(s, index);
                        //System.out.println(index + " " + s.getInputY());
                        index++;
                    }
                    
                    List<DesignSet> outputSets = c.getOutputSets(); 
                    Collections.sort(outputSets, comparator);
                    index = 0;
                    outputSetsOrder.clear();
                    for(DesignSet s : outputSets) {
                        outputSetsOrder.put(s, index);
                        //System.out.println(index + " " + s.getInputY());
                        index++;
                    }
//                    System.out.println("Category " + c.getLabel());
//                    System.out.println("Input sets size: " + c.getInputSets().size());
//                    System.out.println("Output sets size: " + outputSets.size());
//                    System.out.println("Inside sets size: " + c.getInsideSets().size());
                    int[][] insideSets = new int[c.getInputSets().size()][outputSets.size()];
                    Collections.sort(c.getInsideSets(), new Comparator<InsideDesignSet>() {
                        @Override
                        public int compare(InsideDesignSet id1, InsideDesignSet id2) {
                            int i1 = outputSetsOrder.get(id1.getOutputSet());
                            int i2 = outputSetsOrder.get(id2.getOutputSet());
                            if(i1 == i2) {
                                i1 = inputSetsOrder.get(id1.getInputSet());
                                i2 = inputSetsOrder.get(id2.getInputSet());
                            }
                            return i1 - i2;
                            
                        }
                    });
                    for(InsideDesignSet set : c.getInsideSets()) {
                        int insideSetCrossings = 0;
                        DesignSet outputSet = set.getOutputSet();
                        int outputIndex = outputSetsOrder.get(outputSet);
                        DesignSet inputSet = set.getInputSet();
                        int inputIndex = inputSetsOrder.get(inputSet);
//                        System.out.print("input index " + inputIndex + " | ");
//                        System.out.print("output index " + outputIndex);

                        for(int o = outputIndex-1; o >= 0; o--) {
                            for(int i = inputIndex+1; i < insideSets.length; i++) {
                                insideSetCrossings += insideSets[i][o];
                            }
                        }
                        insideSets[inputIndex][outputIndex]++;
                        //System.out.println(" crossings " + insideSetCrossings);
                        insideCrossings += insideSetCrossings;
                    }
                
                    //System.out.println(insideCrossings + " inside crossings for category " + c.getLabel() + " on axis " + axis.getName());
                    if(insideCrossings > maxCrossings) maxCrossings = insideCrossings;
                }
            }
            
        }
        return maxCrossings;
    }

    private void updateCategories() {
        for(ParallelSetsAxis axis : getDisplayAxes()) {
            axis.updateCategories(coordAxisScaleMode);
        }
        filterAxis.updateCategories(coordAxisScaleMode);
        updateCategoriesRequired = false;
    }

    private void updateDesigns(){
        //Update designs in categories
        for (ParallelSetsAxis axis : getDisplayAxes()) {
            for(AxisCategory cat : axis.getCategories()) {
                cat.getDesigns().clear();
            }
            
            for(int i = 0; i < dataSheet.getDesignCount(); i++) {
                Design d = dataSheet.getDesign(i);
                if(!d.isActive()) continue;
                for(AxisCategory cat : axis.getCategories()) {
                    if(!cat.isActive()) continue;
                    AxisCategoryCondition c = cat.getCondition();
                    if(c.test(d)){
                        // design is in category
                        cat.getDesigns().add(d);
                    }
                }
            }
        }
        updateDesignsRequired = false;
    }

    public void zoomIn() {
        if(!filters.isEmpty()) {
            DataSheet ds = getDataSheet();
            long inactivatedDesigns = 0L;
            for(Design d : ds.getDesigns()) {
                if(!d.isActive()) continue;
                boolean inFilter = false;
                for(ParallelSetsFilter filter : filters) {
                    if(filter.test(d)) {
                        inFilter = true;
                        break;
                    }
                }
                if(!inFilter) {
                    d.setActive(false);
                    inactivatedDesigns++;
                }
            }
            long inactiveDesigns = ds.getInactiveDesigns() + inactivatedDesigns;
            ds.setInactiveDesigns(inactiveDesigns);

            updateAxesRequired = true;
            updateCategoriesRequired = true;
            buildDesignSetsRequired = true;
            updateDesignSetsRequired = true;
            
            update();
        }
    }
    
    
    public void zoomOut() {
        for(Design d : getDataSheet().getDesigns()) {
            d.setActive(true);
        }
        this.getDataSheet().setInactiveDesigns(0);
        for(ParallelSetsAxis axis : getDisplayAxes()) {
            for(AxisCategory c : axis.getCategories()) {
                c.active = true;
            }
        }
        
        updateAxesRequired = true;
        updateCategoriesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
        
        update();
    }

    public void copyFilterAxisToChart(){
        // build axis
        ParallelSetsAxisWithoutParameter newAxis = new ParallelSetsAxisWithoutParameter(dataSheet, this, "Axis from Filters");
        parallelSetsAxes.add(newAxis);
        buildDisplayAxes();

        // build categories
        List<AxisCategory> categories = new ArrayList<>();
        AxisCategoryOrCondition previousConditions = new AxisCategoryOrCondition();
        for(AxisCategory category : filterAxis.getCategories()){
            AxisCategoryCondition copyOfCondition = category.getCondition().clone();
            AxisCategoryCondition newCondition = copyOfCondition;
            if(previousConditions.size() > 0) {
                AxisCategoryAndCondition andCondition = new AxisCategoryAndCondition(copyOfCondition, new AxisCategoryNotCondition(previousConditions.clone()));
                andCondition.setLabel(copyOfCondition.getLabel());
                newCondition = andCondition;
            }
            AxisCategory newCategory = new AxisCategory(newCondition, category.getHeightRatio(), category.getDesigns(), newAxis);
            categories.add(newCategory);
            previousConditions.addCondition(copyOfCondition);
        }
        AxisCategory noFilterCategory = filterAxis.getNoFilterCategory();
        categories.add(new AxisCategory(
                noFilterCategory.getCondition().clone(),
                noFilterCategory.getHeightRatio(),
                noFilterCategory.getDesigns(),
                newAxis
        ));
        newAxis.setCategories(categories);

        updateAxesRequired = true;
        updateCategoriesRequired = true;
        buildDesignSetsRequired = true;
        updateDesignSetsRequired = true;
    }

    @Override
    public void setVerticallyOffsetAxisLabels(boolean verticallyOffsetAxisLabels) {
        super.setVerticallyOffsetAxisLabels(verticallyOffsetAxisLabels);

        updateAxesRequired = true;
        updateDesignSetsRequired = true;
    }
}
