
package org.xdat.gui.panels;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.font.FontRenderContext;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.Comparator;
import java.util.List;

import org.xdat.Main;
import org.xdat.chart.*;
import org.xdat.data.DataSheet;
import org.xdat.data.Design;
import org.xdat.data.Parameter;
import org.xdat.exceptions.NoAxisFoundException;
import org.xdat.gui.frames.ChartFrame;
import org.xdat.gui.menus.parallelSetsChart.ParallelSetsContextMenu;
import org.xdat.gui.tables.DataSheetTableColumnModel;

/**
 * Panel that is used to display a parallel sets chart.
 *
 * @author Martin Janda
 */
public class ParallelSetsChartPanel extends ChartPanel implements MouseMotionListener, MouseListener, MouseWheelListener {
    
	static final long serialVersionUID = 0005;
	static final boolean printLog = true;
        static final Cursor handCursor = new Cursor(Cursor.HAND_CURSOR);
        static final Cursor defaultCursor = new Cursor(Cursor.DEFAULT_CURSOR);
        static final Cursor nResizeCursor = new Cursor(Cursor.N_RESIZE_CURSOR);
        static final Cursor eResizeCursor = new Cursor(Cursor.E_RESIZE_CURSOR);
        
	private Main mainWindow;

	private ChartFrame chartFrame;

	private ParallelSetsChart chart;
        
        private BufferedImage canvas;
	private BufferedImage bufferedImage;

	private int dragStartX;

	private int dragStartY;

	private int dragCurrentX;

	private int dragOffsetY;

	private Axis draggedAxis;

    private AxisCategory draggedCategory;
    private ParallelSetsAxis categoryAxis;
    private int dragDeltaY;
    private boolean drawAggregatedValues = false;
        
    private FontRenderContext fontContext;

	
	public ParallelSetsChartPanel(Main mainWindow, ChartFrame chartFrame, ParallelSetsChart chart) {
            super(mainWindow.getDataSheet(), chart);
            this.mainWindow = mainWindow;
            this.chartFrame = chartFrame;
            this.chart = chart;
            log("constructor called");
            this.addMouseListener(this);
            this.addMouseMotionListener(this);
            this.addMouseWheelListener(this);
	}
        
        

        public void setLayoutMode(int mode){
            chart.setLayoutMode(mode);
            updateLayout();
            repaint();
        }

        public int getLayoutMode() {
            return chart.getLayoutMode();
        }
        
        
        @Override
	public void paintComponent(Graphics g) {
            long begin = System.currentTimeMillis();

            int chartWidth = Math.max(chart.getWidth(), this.getWidth());
            if(canvas == null) {
                canvas = new BufferedImage(chartWidth, this.getHeight(), BufferedImage.TYPE_INT_ARGB);
            }
            else if(canvas.getWidth() != chartWidth) {
                canvas = new BufferedImage(chartWidth, this.getHeight(), BufferedImage.TYPE_INT_ARGB);
            }

            if(this.chart.isAntiAliasing()){
                ((Graphics2D) g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            }
            
            g.setColor(Color.WHITE);
            g.clearRect(0, 0, getWidth(), getHeight());

            if (this.draggedAxis != null) {
                g.drawImage(this.canvas, 0, 0, null);
                //log("paintComponent: drawing dragged axis ");
                g.setColor(new Color(100, 100, 100));
                int yPosition = chart.getAxisTopPos();
                int xPosition = this.dragCurrentX;
                g.drawLine(xPosition - 1, yPosition, xPosition - 1, yPosition + (chart.getAxisHeight()));
                g.drawLine(xPosition, yPosition, xPosition, yPosition + (chart.getAxisHeight()));
                g.drawLine(xPosition + 1, yPosition, xPosition + 1, yPosition + (chart.getAxisHeight()));
            } 
            else if (draggedCategory != null) {
                g.drawImage(this.canvas, 0, 0, null);
                g.setColor(new Color(100, 100, 100, 125));

                int[] pointsX = {
                    categoryAxis.getPosition().x,
                    categoryAxis.getPosition().x + categoryAxis.getBoxWidth(),
                    categoryAxis.getPosition().x + categoryAxis.getBoxWidth(),
                    categoryAxis.getPosition().x
                };
                int[] pointsY = {
                    draggedCategory.getyUpPosition() + dragDeltaY,
                    draggedCategory.getyUpPosition() + dragDeltaY,
                    draggedCategory.getyUpPosition() + draggedCategory.getHeight() + dragDeltaY,
                    draggedCategory.getyUpPosition() + draggedCategory.getHeight() + dragDeltaY
                };
                g.fillPolygon(pointsX, pointsY, 4);
            } 
            else {
                Graphics2D cg = (Graphics2D)canvas.getGraphics();
                if(this.chart.isAntiAliasing()){
                    cg.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
                }
                cg.setBackground(Color.WHITE);
                cg.clearRect(0, 0, canvas.getWidth(), canvas.getHeight());
                this.drawDesigns(cg);
                this.drawAxes(cg);
                g.drawImage(canvas, 0, 0, null);
            }
            //log("Painting took "+(System.currentTimeMillis()-begin)+" ms");
	}
	
	public ParallelSetsAxis withinAxisDragZone(int x, int y) {
            for (ParallelSetsAxis axis : chart.getDisplayAxes()) {
                if(!axis.isActive()) continue;
                int axisX = axis.getPosition().x;
                int axisY = axis.getPosition().y;
                
                Rectangle2D stringBounds = axis.getLabelFont().getStringBounds(axis.getName(), axis.fontContext);

                int labelY = axisY - ParallelSetsAxis.LABEL_HEIGHT - axis.getYLabelOffset();
                
                if (stringBounds.contains(x - axisX, y - labelY)) {
                    return axis;
                }
            }
            return null;
	}

        public ParallelSetsAxis onAxisSortButton(int x, int y){
            if(!chart.isSortButtonOn()) return null;
            for (ParallelSetsAxis axis : chart.getDisplayAxes()) {
                if(axis.getParameter() == null || !axis.getParameter().isNumeric()) continue;
                
                Rectangle2D stringBounds = axis.getLabelFont().getStringBounds("↑", axis.fontContext);
                int buttonX = axis.getPosition().x - 15;
                int buttonY = axis.getPosition().y - ParallelSetsAxis.LABEL_HEIGHT - axis.getYLabelOffset();
                if (stringBounds.contains(x - buttonX, y - buttonY)) {
                    return axis;
                }
            }
            return null;
        }

        @Override
        public void updateLayout(){
            log("updateLayout");
            chart.update();
            int width = chart.getWidth();
            canvas = new BufferedImage(width, this.getHeight(), BufferedImage.TYPE_INT_ARGB);
            bufferedImage = new BufferedImage(width, this.getHeight(), BufferedImage.TYPE_INT_ARGB);
            Dimension d = new Dimension(width, this.getHeight());
            this.setMinimumSize(d);
            this.setPreferredSize(d);
            this.setMaximumSize(d);
            chartFrame.revalidate();
        }

	public void drawDesigns(final Graphics g) {
            if(chart.isAntiAliasing()){
                Graphics2D graphics2D = (Graphics2D) g;
                graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            }

            // Draw parallelograms
            ParallelSetsAxis lastAxis = null;
            for(ParallelSetsAxis currentAxis : chart.getDisplayAxes()) {
                if(lastAxis == null){
                    lastAxis = currentAxis;
                    continue;
                }
                
                Parameter p = lastAxis.getParameter();
                if(lastAxis.isExpanded() && p != null && p.getAxisClone() != lastAxis && !p.getChildren().isEmpty()) {
                    float width = 0f;
                    ParallelSetsAxis lastChild = lastAxis;
                    for(Parameter child : lastAxis.getParameter().getChildren()) {
                        ParallelSetsAxis childAxis = (ParallelSetsAxis) child.getAxis();
                        width += lastChild.getWidth() * 0.5 + childAxis.getWidth() * 0.5 + lastChild.getBoxWidth();
                        lastChild = childAxis;
                    }
                    width += lastChild.getWidth() * 0.5 + currentAxis.getWidth() * 0.5 + lastChild.getBoxWidth();
                    width -= currentAxis.getBoxWidth();
                    g.setColor(new Color(0.95f, 0.95f, 0.95f));
                    g.fillRect(lastAxis.getPosition().x + lastAxis.getBoxWidth(), lastAxis.getPosition().y, (int) width, chart.getAxisHeight()-1);
                }
                
                for(AxisCategory category : currentAxis.getCategories()) {
                    if(!category.isActive()) continue;
                    if(!currentAxis.isCoordinatesAxis()){
                        g.setColor(Color.WHITE);
                        g.fillRect(currentAxis.getPosition().x, category.getyUpPosition(), currentAxis.getBoxWidth(), category.getHeight());
                    }
                    
                    for(DesignSet set : category.getInputSets()) {
                        int[] pointsX = {
                            lastAxis.getPosition().x + lastAxis.getBoxWidth(),
                            currentAxis.getPosition().x,
                            currentAxis.getPosition().x,
                            lastAxis.getPosition().x + lastAxis.getBoxWidth()
                        };
                        int[] pointsY = {
                            (int) set.getInputY(),
                            (int) set.getOutputY(),
                            (int) (set.getOutputY() + set.getOutputHeight()),
                            (int) (set.getInputY() + set.getInputHeight())
                        };

                        Color color = opacityMapping(set);
                        g.setColor(color);
                        if(-pointsY[0] + pointsY[3] <= 1) {
                            g.drawLine(pointsX[0], (int)((pointsY[0] + pointsY[3])*0.5f), pointsX[1], (int)((pointsY[1] + pointsY[2])*0.5f));
                            //log("Drawing line insteadof polygon");
                        }
                        else {
                            g.fillPolygon(pointsX, pointsY, 4);
                        }
                        
                        
                        // DRAW INSIDE AXIS TREE
                        if(!currentAxis.isCoordinatesAxis()){
                            for (InsideDesignSet insideSet : set.getOutputInsideSets()) {
                                int[] pX = {
                                    currentAxis.getPosition().x,
                                    currentAxis.getPosition().x + currentAxis.getBoxWidth(),
                                    currentAxis.getPosition().x + currentAxis.getBoxWidth(),
                                    currentAxis.getPosition().x
                                };
                                int[] pY = {
                                    (int) insideSet.getInputY(),
                                    (int) insideSet.getOutputY(),
                                    (int) (insideSet.getOutputY() + insideSet.getHeight()),
                                    (int) (insideSet.getInputY() + insideSet.getHeight())
                                };

                                Color insideColor = opacityMapping(insideSet);
                                g.setColor(insideColor);
                                if(-pY[0] + pY[3] <= 1) {
                                    g.drawLine(pX[0], (int)((pY[0] + pY[3])*0.5f), pX[1], (int)((pY[1] + pY[2])*0.5f));
                                    //log("Drawing line insteadof polygon");
                                }
                                else {
                                    g.fillPolygon(pX, pY, 4);
                                }

                            }
                        }
                        
                        // Draw designs on coordinates axis
                        
                        if(ParallelSetsChart.NO_FILTER_COLOR.equals(color)) continue; // skip gray sets
                        // First half - designs incoming to axis
                        if(currentAxis.isCoordinatesAxis()){
                            // sort designs
                            List<Design> designs = set.getDesigns();
                            designs.sort(new Comparator<Design>() {
                                @Override
                                public int compare(Design d1, Design d2) {
                                    if(currentAxis.isInverted()) return Double.compare(d1.getDoubleValue(currentAxis.getParameter()), d2.getDoubleValue(currentAxis.getParameter()));
                                    return Double.compare(d2.getDoubleValue(currentAxis.getParameter()), d1.getDoubleValue(currentAxis.getParameter()));
                                }
                            });
                            // draw designs
                            int count = 0;
                            float totalHeight = 0;
                            float height = set.getOutputHeight() / (float) designs.size();
                            float yUp = set.getOutputY();
                            float yAverage = 0f;
                            for (Design design : designs) {

                                double value = design.getDoubleValue(currentAxis.getParameter());
                                if (Double.isNaN(value)) {
                                    value = currentAxis.getMin();
                                }

                                double yOnAxis;
                                if(category.getCondition().getClass() != AxisCategoryStringCondition.class && chart.getCoordAxisScaleMode() == ParallelSetsChart.COORD_AXIS_SCALING_AXIS){
                                    AxisCategoryRangeCondition condition = (AxisCategoryRangeCondition) category.getCondition();
                                    double minValue = condition.getMin();
                                    double maxValue = condition.getMax();
                                    double ratio = (value -  minValue) / (maxValue - minValue);
                                    double posRelToBottom = category.getHeight() * ratio;
                                    if(currentAxis.isInverted()){
                                        yOnAxis = category.getyUpPosition() + posRelToBottom;
                                    }else{
                                        yOnAxis = category.getyUpPosition() + category.getHeight() - posRelToBottom;
                                    }
                                }else{
                                    double ratio = (value - currentAxis.getMin()) / currentAxis.getRange();
                                    double posRelToBottom = chart.getAxisHeight() * ratio;
                                    if(currentAxis.isInverted()){
                                        yOnAxis = currentAxis.getPosition().y + posRelToBottom;
                                    }else{
                                        yOnAxis = currentAxis.getPosition().y + chart.getAxisHeight() - posRelToBottom;
                                    }
                                }
                                yAverage += yOnAxis;

                                if(!drawAggregatedValues) {
                                    float posXOnAxis = currentAxis.getPosition().x + currentAxis.getBoxWidth() * 0.5f;
                                    int[] posX = {
                                        currentAxis.getPosition().x,
                                        (int) posXOnAxis,
                                        currentAxis.getPosition().x
                                    };
                                    int[] posY = {
                                        (int) yUp,
                                        (int) yOnAxis,
                                        (int) (yUp + height)
                                    };
                                    g.setColor(color);
                                    g.fillPolygon(posX, posY, 3);
                                }

                                yUp += height;
                                count++;
                                totalHeight += height;
                                height = (set.getOutputHeight() - totalHeight) / (float)(designs.size() - count);
                            }

                            if(drawAggregatedValues) {
                                float posXOnAxis = currentAxis.getPosition().x + currentAxis.getBoxWidth() * 0.5f;
                                int[] posX = {
                                    currentAxis.getPosition().x,
                                    (int) posXOnAxis,
                                    currentAxis.getPosition().x
                                };
                                int[] posY = {
                                    (int) set.getOutputY(),
                                    (int) (yAverage/(float) count),
                                    (int) (set.getOutputY() + set.getOutputHeight())
                                };
                                g.setColor(color);
                                g.fillPolygon(posX, posY, 3);
                            }

                        }
                        // Second half - designs leaving axis
                        if(lastAxis.isCoordinatesAxis()){
                            // sort designs
                            List<Design> designs = set.getDesigns();
                            final ParallelSetsAxis lAxis = lastAxis;
                            designs.sort(new Comparator<Design>() {
                                @Override
                                public int compare(Design d1, Design d2) {
                                    if(lAxis.isInverted()) return Double.compare(d1.getDoubleValue(lAxis.getParameter()), d2.getDoubleValue(lAxis.getParameter()));
                                    return Double.compare(d2.getDoubleValue(lAxis.getParameter()), d1.getDoubleValue(lAxis.getParameter()));
                                }
                            });
                            // draw designs
                            int count = 0;
                            float totalHeight = 0;
                            float height = set.getInputHeight()/ (float) designs.size();
                            float yUp = set.getInputY();
                            float yAverage = 0f;
                            AxisCategory inpCategory = set.getInputCategory();
                            for (Design design : designs) {

                                double value = design.getDoubleValue(lastAxis.getParameter());
                                if (Double.isNaN(value)) {
                                    value = lastAxis.getMin();
                                }

                                double yOnAxis;
                                if(inpCategory.getCondition().getClass() != AxisCategoryStringCondition.class && chart.getCoordAxisScaleMode() == ParallelSetsChart.COORD_AXIS_SCALING_AXIS){
                                    AxisCategoryRangeCondition condition = (AxisCategoryRangeCondition) inpCategory.getCondition();
                                    double minValue = condition.getMin();
                                    double maxValue = condition.getMax();
                                    double ratio = (value -  minValue) / (maxValue - minValue);
                                    double posRelToBottom = inpCategory.getHeight() * ratio;
                                    if(lastAxis.isInverted()){
                                        yOnAxis = inpCategory.getyUpPosition() + posRelToBottom;
                                    }else{
                                        yOnAxis = inpCategory.getyUpPosition() + inpCategory.getHeight() - posRelToBottom;
                                    }
                                }else{
                                    double ratio = (value - lastAxis.getMin()) / lastAxis.getRange();
                                    double posRelToBottom = chart.getAxisHeight() * ratio;
                                    if(lastAxis.isInverted()){
                                        yOnAxis = lastAxis.getPosition().y + posRelToBottom;
                                    }else{
                                        yOnAxis = lastAxis.getPosition().y + chart.getAxisHeight() - posRelToBottom;
                                    }
                                }
                                yAverage += yOnAxis;

                                if(!drawAggregatedValues) {
                                    float posXOnAxis = lastAxis.getPosition().x + lastAxis.getBoxWidth()*0.5f;
                                    int[] posX = {
                                        lastAxis.getPosition().x + lastAxis.getBoxWidth(),
                                        (int) posXOnAxis,
                                        lastAxis.getPosition().x + lastAxis.getBoxWidth()
                                    };
                                    int[] posY = {
                                        (int) yUp,
                                        (int) yOnAxis,
                                        (int) (yUp + height)
                                    };
                                    g.setColor(color);
                                    g.fillPolygon(posX, posY, 3);
                                }

                                yUp += height;
                                count++;
                                totalHeight += height;
                                height = (set.getInputHeight() - totalHeight) / (designs.size() - count);
                            }

                            if(drawAggregatedValues) {
                                float posXOnAxis = lastAxis.getPosition().x + lastAxis.getBoxWidth()*0.5f;
                                int[] posX = {
                                    lastAxis.getPosition().x + lastAxis.getBoxWidth(),
                                    (int) posXOnAxis,
                                    lastAxis.getPosition().x + lastAxis.getBoxWidth()
                                };
                                int[] posY = {
                                    (int) set.getInputY(),
                                    (int) (yAverage / (float) count),
                                    (int) (set.getInputY() + set.getInputHeight())
                                };
                                g.setColor(color);
                                g.fillPolygon(posX, posY, 3);
                            }
                        }
                    }
                }
                lastAxis = currentAxis;
            }
	}
	
	public void drawAxes(Graphics g) {
            Graphics2D g2 = (Graphics2D) g;
            for (ParallelSetsAxis currentAxis : chart.getDisplayAxes()) {
                currentAxis.drawAxis(g2);
            }
	}

	private ParallelSetsAxis getAxisAtLocation(int x) {
        for (ParallelSetsAxis axis : chart.getDisplayAxes()) {
            if(x >= axis.getPosition().x - 0.5 * axis.getWidth() + axis.getBoxWidth() / 2
                    && x < axis.getPosition().x + 0.5 * axis.getWidth() + axis.getBoxWidth() / 2) {
                return axis;
            }
        }
        ParallelSetsAxis axis = chart.getFilterAxis();
        if(x >= axis.getPosition().x - 0.5 * axis.getWidth() + axis.getBoxWidth() / 2
                && x < axis.getPosition().x + 0.5 * axis.getWidth() + axis.getBoxWidth() / 2) {
            return axis;
        }
        return null;
	}
        
    private int getParameterIndex(Axis axis) {
        DataSheet ds = chart.getDataSheet();
        int currentIndex = 0;
        for(int i = 0; i < ds.getParameterCount(); i++) {
            Parameter p = ds.getParameter(i);
            if(p == axis.getParameter()) break;
            currentIndex++;
        }
        return currentIndex;
    }

	private int getNewParameterIndex(int x, ParallelSetsAxis dragedAxis) throws NoAxisFoundException {
            ParallelSetsAxis axisAfter = null;
            Parameter draggedParameter = draggedAxis.getParameter();
            if(draggedParameter == null) return 0;
            boolean after = true;
            for (ParallelSetsAxis axis : chart.getDisplayAxes()) {
                Parameter p = axis.getParameter();
                if(p == null) continue;
                if(draggedParameter.isAggregated()) {
                    if(!p.isAggregated()) continue;
                    if(p.getParent() != draggedParameter.getParent()) continue;
                }
                else {
                    if(p.isAggregated()) continue;
                }
                if(p.getAxisClone() == axis) continue;
                if(x < axis.getPosition().x) {
                    axisAfter = axis;
                    break;
                }
            }
            if(axisAfter == null) {
                if(draggedParameter.isAggregated()) {
                    List<Parameter> children = draggedParameter.getParent().getChildren();
                    Parameter child = children.get(children.size()-1);
                    axisAfter = (ParallelSetsAxis) child.getAxis();
                }
                else {
                    List<ParallelSetsAxis> list = chart.getDisplayAxes();
                    axisAfter = list.get(list.size()-1);
                }
                after = false;
            }
            int parameterIndex = getParameterIndex(axisAfter);
            if(after) {
                parameterIndex--;
            }
            return parameterIndex;
	}

    private int getAxisIndex(Axis axis) {
        return chart.getAllAxes().indexOf(axis);
    }

    private int getNewAxisIndex(int x, ParallelSetsAxis dragedAxis) throws NoAxisFoundException {

        ParallelSetsAxis axisAfter = chart.getAllAxes().get(chart.getAllAxes().size() - 1);
        boolean after = false;
        for(ParallelSetsAxis axis : chart.getAllAxes()) {
            if(x < axis.getPosition().x) {
                axisAfter = axis;
                after = true;
                break;
            }
        }
        int axisIndex = getAxisIndex(axisAfter);
        if(after) {
            axisIndex--;
        }
        return axisIndex;
    }

	public void mouseClicked(MouseEvent e) {
            if(e.getButton() == 1){
                int x = e.getX();
                int y = e.getY();
                
                ParallelSetsAxis axis = getAxisAtLocation(x);
                if(axis != null){
                    AxisCategory category = axis.getCategoryInDragZone(y);
                    if(category != null) {
                        log("Clicked on category " + category.getLabel());
                        if(chart.getOneAxis()){
                            if(axis == chart.getSeparatingAxis()){
                                if(chart.getLayoutMode() == ParallelSetsChart.TREE_LAYOUT){
                                    chart.setHighlightedCategory(category);
                                }else{
                                    chart.setOneAxis(false);
                                    chart.setSelectedFilter(null);
                                    chart.getFilters().clear();
                                    chart.clearSelectedCategories();
                                    chart.addSelectedCategory(category);
                                }
                            }else{
                                log("Clicked on axis " + axis.getName());
                                chart.setSeparatingAxis(axis);
                            }
                            chartFrame.getSidePanel().updateData();
                        }else{
                            chart.switchSelectedCategory(category);
                        }
                        updateLayout();
                        chartFrame.repaint();
                    } else {
                        axis = withinAxisDragZone(x, y);
                        if(axis != null) {
                            chart.setOneAxis(true);
                            log("Clicked on axis " + axis.getName());
                            chart.setSeparatingAxis(axis);
                            chartFrame.getSidePanel().updateData();
                            updateLayout();
                            repaint();
                        }
                    }
                }
            }
            else if (e.getButton() == 3) {
                log("mouseClicked: button " + e.getButton());
                int x = e.getX();
                int y = e.getY();

                log("mouseClicked: x " + x);
                log("mouseClicked: y " + y);
                ParallelSetsAxis axis = withinAxisDragZone(x, y);
                if(axis != null) {
                    log("Clicked on axis " + axis.getName());
                    (new ParallelSetsContextMenu(this.mainWindow, this.chartFrame, axis)).show(this, x, y);
                }
            }

	}

	
	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mousePressed(MouseEvent e) {
            dragStartX = e.getX();
            dragStartY = e.getY();
            dragDeltaY = 0;
            
            this.storeBufferedImage();
            if ( e.getButton() == 1) {
                ParallelSetsAxis axis = getAxisAtLocation(dragStartX);
                if(axis != null) {
                    AxisCategory cat = axis.getCategoryInDragZone(dragStartY);

                    // DRAG CATEGORY
                    if(cat != null){
                        this.categoryAxis = getAxisAtLocation(dragStartX);
                        if(categoryAxis.getParameter() == null || !categoryAxis.getParameter().isNumeric()){
                            this.draggedCategory = cat;
                            setCursor(nResizeCursor);
                            log("mousePressed: Drag started, dragged category : " + categoryAxis.getName() + ", index: " + categoryAxis.getCategories().indexOf(draggedCategory));
                        }
                        // DRAG AXIS
                    } else {
                        axis = withinAxisDragZone(dragStartX, dragStartY);
                        if(axis != null){
                            this.draggedAxis = axis;
                            setCursor(eResizeCursor);
                            log("mousePressed: Drag started, dragged axis : " + this.draggedAxis.getName());
                        }
                    }
                }
                // click to sort button
                if(chart.isSortButtonOn()){
                    ParallelSetsAxis axisToSort = onAxisSortButton(dragStartX, dragStartY);
                    if(axisToSort != null){
                        chart.invertAxis(axisToSort);
                        updateLayout();
                        repaint();
                    }
                }
            }
	}
	
	public void mouseReleased(MouseEvent e) {
            setCursor(defaultCursor);

            if(dragStartX == e.getX() && dragStartY == e.getY()) {
                draggedAxis = null;
                draggedCategory = null;
                return;
            }

            boolean repaintRequired = false;
            if (this.draggedAxis != null) {
                repaintRequired = true;
                try {
                    int oldIndex = getAxisIndex(draggedAxis);
                    int newIndex = getNewAxisIndex(e.getX(), (ParallelSetsAxis) draggedAxis);
                    int oldPIndex = getParameterIndex(draggedAxis);
                    int newPIndex = getNewParameterIndex(e.getX(), (ParallelSetsAxis) draggedAxis);
                    log("mouseReleased: dragged param " + this.draggedAxis.getName() + " had index " + oldIndex);
                    log("mouseReleased: dragged param " + this.draggedAxis.getName() + " has new index " + newIndex);
                    DataSheetTableColumnModel cm = (DataSheetTableColumnModel) this.mainWindow.getDataSheetTablePanel().getDataTable().getColumnModel();
                    if (newIndex > oldIndex) {
                        chart.moveAxisAndParameter(oldIndex, newIndex, oldPIndex, newPIndex, cm);
                    } else if (newIndex < oldIndex) {
                        chart.moveAxisAndParameter(oldIndex, newIndex+1, oldPIndex, newPIndex+1, cm);
                    }
                } catch (NoAxisFoundException e1) {
                    e1.printStackTrace();
                }
                this.draggedAxis = null;
            }
            else if(this.draggedCategory != null){
                repaintRequired = true;
                //ParallelSetsAxis axis = getAxisAtLocation(e.getX());
                AxisCategory cat = categoryAxis.getCategoryInDragZone(e.getY());
                if(cat != null){
                    //log("mouseReleased: dragged axis " + this.draggedAxis.getName() + " had index " + currentIndex);
                    int newIndex = categoryAxis.getCategories().indexOf(cat);
                    chart.moveCategory(draggedCategory, newIndex);
                }
                draggedCategory = null;
                categoryAxis = null;
            }
            if (repaintRequired) {
                updateLayout();
                chartFrame.getSidePanel().updateData();
                this.mainWindow.repaintAllChartFrames();
            }
	}

	public void mouseMoved(MouseEvent e) {
            boolean needRepaint = false;
            
            int x = e.getX();
            int y = e.getY();

            if (withinAxisDragZone(x, y) != null) {
                setCursor(handCursor);
            } else if (onAxisSortButton(x, y) != null){
                setCursor(handCursor);
            } else {
                setCursor(defaultCursor);
            }

            if (needRepaint) {
                this.repaint();
            }
	}

	public void mouseDragged(MouseEvent e) {
            if (this.draggedAxis != null) {
                this.dragCurrentX = e.getX();
                this.repaint();
            } else if (draggedCategory != null){
                dragDeltaY = e.getY() - dragStartY;
                repaint();
            }
	}

	@Override
	public void mouseWheelMoved(MouseWheelEvent e) {
		int modifier = e.getModifiers();
		if (modifier == 0) {
			this.chart.incrementAxisWidth(-e.getUnitsToScroll());
		} else if (modifier == MouseWheelEvent.CTRL_MASK) {
			int x = e.getX();
                        Axis axis = this.getAxisAtLocation(x);
                        if(axis != null) {
                            log("wheeled on axis " + axis.getName());
                            axis.setWidth(Math.max(0, axis.getWidth() - e.getUnitsToScroll()));
                        }
		}
                else if(modifier == MouseWheelEvent.SHIFT_MASK) {
                    this.chart.incrementAxisBoxWidth(-e.getUnitsToScroll());
                }
		else if (modifier == MouseWheelEvent.ALT_MASK) {
			int x = e.getX();
                        Axis axis = this.getAxisAtLocation(x);
                        if(axis != null) {
                            log("wheeled on axis " + axis.getName());
                            axis.setTicCount(Math.max(2, axis.getTicCount() - e.getWheelRotation()));
                        }
		}
                chart.updateAxes();
                this.setBounds(0, 0, chart.getWidth(), getHeight());
                chartFrame.repaint();
	}

	
	public void storeBufferedImage() {
		this.bufferedImage = new BufferedImage(chartFrame.getChartPanel().getWidth(), chartFrame.getChartPanel().getHeight(), BufferedImage.TYPE_INT_ARGB);
		Graphics g = this.bufferedImage.createGraphics();
		this.paintComponent(g);
		g.dispose();
	}

	
	private void log(String message) {
		if (ParallelSetsChartPanel.printLog && Main.isLoggingEnabled()) {
			System.out.println(this.getClass().getName() + "." + message);
		}
	}

    private Color opacityMapping(DesignSet set) {
       Color setColor = set.getColor();
//        FilterAxisCategory inFilter = null;
//        for(AxisCategory c : chart.getFilterAxis().getCategories()) {
//            FilterAxisCategory fc = (FilterAxisCategory) c;
//            ParallelSetsFilter filter = fc.getFilter();
//            if(filter.getColor().equals(setColor)) {
//                inFilter = fc;
//                break;
//            }
//        }
        float[] sc = setColor.getComponents(new float[4]);
        float alphaScale = 1f;
//        if(inFilter != null) {
//            alphaScale = (float) set.getDesigns().size() / (float) inFilter.getDesigns().size();
//        }
//        alphaScale = (float) Math.pow(alphaScale, 0.3);
        return new Color(sc[0], sc[1], sc[2], sc[3]*alphaScale);
    }
    
    private Color opacityMapping(InsideDesignSet set) {
        Color setColor = set.getColor();
//        FilterAxisCategory inFilter = null;
//        for(AxisCategory c : chart.getFilterAxis().getCategories()) {
//            FilterAxisCategory fc = (FilterAxisCategory) c;
//            ParallelSetsFilter filter = fc.getFilter();
//            if(filter.getColor().equals(setColor)) {
//                inFilter = fc;
//                break;
//            }
//        }
        float[] sc = setColor.getComponents(new float[4]);
        float alphaScale = 1f;
//        if(inFilter != null) {
//            alphaScale = (float) set.getDesigns().size() / (float) inFilter.getDesigns().size();
//        }
//        alphaScale = (float) Math.pow(alphaScale, 0.3);
        return new Color(sc[0], sc[1], sc[2], sc[3]*alphaScale);
    }

    /**
     * @return the drawAggregatedValues
     */
    public boolean getDrawAggregatedValues() {
        return drawAggregatedValues;
    }

    /**
     * @param drawAggregatedValues the drawAggregatedValues to set
     */
    public void setDrawAggregatedValues(boolean drawAggregatedValues) {
        this.drawAggregatedValues = drawAggregatedValues;
    }
        
        

}
